# Copyright 2021 The NPLinker Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# CG: these functions are not used by other, could be removed
import logging
from . import nplinker_iokr
from .spectrum import MSSpectrum


logger = logging.getLogger(__name__)


def run_iokr_ranking(spec, bgc_list):
    logger.debug(f"IOKR datapath: {nplinker_iokr.get_datapath()}")

    # extract SMILES strings for each supplied BGC (caching them in the object)
    smiles = [bgc.smiles for bgc in bgc_list]

    # TODO could probably handle this more gracefully by not submitting any BGC with no SMILES
    if None in smiles:
        logger.error("Failed to run IOKR scoring, one or more BGCs missing SMILES data")
        return None

    iokr_server = nplinker_iokr.get_iokr_server()
    rank = iokr_server.rank_smiles(MSSpectrum(spec=spec), smiles)
    return [bgc_list[i] for i in rank]


def run_iokr_scoring(spec_list, bgc_list):
    logger.debug(f"IOKR datapath: {nplinker_iokr.get_datapath()}")
    smiles = [bgc.smiles for bgc in bgc_list]
    if None in smiles:
        logger.error("Failed to run IOKR scoring, one or more BGCs missing SMILES data")
        return None

    spectra = [MSSpectrum(spec=spec) for spec in spec_list]

    iokr_server = nplinker_iokr.get_iokr_server()
    rank = iokr_server.score_smiles(spectra, smiles)

    return rank



from __future__ import annotations
import bz2
import csv
import functools
import gzip
import hashlib
import logging
import lzma
import os
import os.path
import shutil
import sys
import tarfile
import warnings
import zipfile
from collections.abc import Callable
from collections.abc import Sequence
from os import PathLike
from pathlib import Path
from typing import IO
import httpx
from rich.progress import BarColumn
from rich.progress import DownloadColumn
from rich.progress import Progress
from rich.progress import TextColumn
from rich.progress import TimeElapsedColumn
from rich.progress import TimeRemainingColumn
from rich.progress import TransferSpeedColumn


logger = logging.getLogger(__name__)

#
# Decorators
#


def check_disk_space(func):
    """A decorator to check available disk space.

    If the available disk space is less than 500GB, raise and log a warning.

    Warnings:
        UserWarning: If the available disk space is less than 500GB.
    """

    @functools.wraps(func)
    def wrapper_check_disk_space(*args, **kwargs):
        _, _, free = shutil.disk_usage("/")
        free_gb = free // (2**30)
        if free_gb < 50:
            warning_message = f"Available disk space is {free_gb}GB. Is it enough for your project?"
            logger.warning(warning_message)
            warnings.warn(warning_message, UserWarning)
        return func(*args, **kwargs)

    return wrapper_check_disk_space


#
# Other utility functions
#


def is_file_format(file: str | PathLike, format: str = "tsv") -> bool:
    """Check if the file is in the given format.

    Args:
        file: Path to the file to check.
        format: The format to check for, either "tsv" or "csv".

    Returns:
        True if the file is in the given format, False otherwise.
    """
    try:
        with open(file, "rt") as f:
            if format == "tsv":
                reader = csv.reader(f, delimiter="\t")
            elif format == "csv":
                reader = csv.reader(f, delimiter=",")
            else:
                raise ValueError(f"Unknown format '{format}'.")
            for _ in reader:
                pass
        return True
    except csv.Error:
        return False


# Functions below are adapted from torchvision library,
# see: https://github.com/pytorch/vision/blob/main/torchvision/datasets/utils.py.
#
# BSD 3-Clause License
# Copyright (c) Soumith Chintala 2016
# You may obtain a copy of the License at
#    https://github.com/pytorch/vision/blob/main/LICENSE


def calculate_md5(fpath: str | PathLike, chunk_size: int = 1024 * 1024) -> str:
    """Calculate the MD5 checksum of a file.

    Args:
        fpath: Path to the file.
        chunk_size: Chunk size for reading the file. Defaults to 1024*1024.

    Returns:
        MD5 checksum of the file.
    """
    if sys.version_info >= (3, 9):
        md5 = hashlib.md5(usedforsecurity=False)
    else:
        md5 = hashlib.md5()
    with open(fpath, "rb") as f:
        for chunk in iter(lambda: f.read(chunk_size), b""):
            md5.update(chunk)
    return md5.hexdigest()


def check_md5(fpath: str | PathLike, md5: str) -> bool:
    """Verify the MD5 checksum of a file.

    Args:
        fpath: Path to the file.
        md5: MD5 checksum to verify.

    Returns:
        True if the MD5 checksum matches, False otherwise.
    """
    return md5 == calculate_md5(fpath)


@check_disk_space
def download_url(
    url: str,
    root: str | PathLike,
    filename: str | None = None,
    md5: str | None = None,
    http_method: str = "GET",
    allow_http_redirect: bool = True,
) -> None:
    """Download a file from a url and place it in root.

    Args:
        url: URL to download file from
        root: Directory to place downloaded file in. If it doesn't exist, it will be created.
        filename: Name to save the file under. If None, use the
            basename of the URL.
        md5: MD5 checksum of the download. If None, do not check.
        http_method: HTTP request method, e.g. "GET", "POST".
            Defaults to "GET".
        allow_http_redirect: If true, enable following redirects for all HTTP ("http:") methods.
    """
    root = transform_to_full_path(root)
    # create the download directory if not exist
    root.mkdir(exist_ok=True)
    if not filename:
        filename = Path(url).name
    fpath = root / filename

    # check if file is already present locally
    if fpath.is_file() and md5 is not None and check_md5(fpath, md5):
        logger.info("Using downloaded and verified file: " + str(fpath))
        return

    # download the file
    logger.info(f"Downloading {filename} to {root}")
    with open(fpath, "wb") as fh:
        with httpx.stream(http_method, url, follow_redirects=allow_http_redirect) as response:
            if not response.is_success:
                fpath.unlink(missing_ok=True)
                raise RuntimeError(
                    f"Failed to download url {url} with status code {response.status_code}"
                )
            total = int(response.headers.get("Content-Length", 0))

            with Progress(
                TextColumn("[progress.description]{task.description}"),
                BarColumn(bar_width=None),
                "[progress.percentage]{task.percentage:>3.1f}%",
                "•",
                DownloadColumn(),
                "•",
                TransferSpeedColumn(),
                "•",
                TimeRemainingColumn(),
                "•",
                TimeElapsedColumn(),
            ) as progress:
                task = progress.add_task(f"[hot_pink]Downloading {fpath.name}", total=total)
                for chunk in response.iter_bytes():
                    fh.write(chunk)
                    progress.update(task, advance=len(chunk))

    # check integrity of downloaded file
    if md5 is not None and not check_md5(fpath, md5):
        raise RuntimeError("MD5 validation failed.")


def list_dirs(root: str | PathLike, keep_parent: bool = True) -> list[str]:
    """List all directories at a given root.

    Args:
        root: Path to directory whose folders need to be listed
        keep_parent: If true, prepends the path to each result, otherwise
            only returns the name of the directories found
    """
    root = transform_to_full_path(root)
    directories = [str(p) for p in root.iterdir() if p.is_dir()]
    if not keep_parent:
        directories = [os.path.basename(d) for d in directories]
    return directories


def list_files(
    root: str | PathLike,
    prefix: str | tuple[str, ...] = "",
    suffix: str | tuple[str, ...] = "",
    keep_parent: bool = True,
) -> list[str]:
    """List all files at a given root.

    Args:
        root: Path to directory whose files need to be listed
        prefix: Prefix of the file names to match,
            Defaults to empty string '""'.
        suffix: Suffix of the files to match, e.g. ".png" or
            (".jpg", ".png").
            Defaults to empty string '""'.
        keep_parent: If true, prepends the parent path to each
            result, otherwise only returns the name of the files found.
            Defaults to False.
    """
    root = Path(root)
    files = [
        str(p)
        for p in root.iterdir()
        if p.is_file() and p.name.startswith(prefix) and p.name.endswith(suffix)
    ]

    if not keep_parent:
        files = [os.path.basename(f) for f in files]

    return files


def _extract_tar(
    from_path: str | PathLike,
    to_path: str | PathLike,
    members: Sequence[tarfile.TarInfo] | None,
    compression: str | None,
) -> None:
    with tarfile.open(from_path, f"r:{compression[1:]}" if compression else "r") as tar:
        tar.extractall(to_path, members)


_ZIP_COMPRESSION_MAP: dict[str, int] = {
    ".bz2": zipfile.ZIP_BZIP2,
    ".xz": zipfile.ZIP_LZMA,
}


def _extract_zip(
    from_path: str | PathLike,
    to_path: str | PathLike,
    members: Sequence[str | zipfile.ZipInfo] | None,
    compression: str | None,
) -> None:
    with zipfile.ZipFile(
        from_path,
        "r",
        compression=_ZIP_COMPRESSION_MAP[compression] if compression else zipfile.ZIP_STORED,
    ) as zf:
        zf.extractall(to_path, members)


_ARCHIVE_EXTRACTORS: dict[str, Callable[[str, str, list | None, str | None], None]] = {
    ".tar": _extract_tar,
    ".zip": _extract_zip,
}
_COMPRESSED_FILE_OPENERS: dict[str, Callable[..., IO]] = {
    ".bz2": bz2.open,
    ".gz": gzip.open,
    ".xz": lzma.open,
}
_FILE_TYPE_ALIASES: dict[str, tuple[str | None, str | None]] = {
    ".tbz": (".tar", ".bz2"),
    ".tbz2": (".tar", ".bz2"),
    ".tgz": (".tar", ".gz"),
}


def _detect_file_type(file: str | Path) -> tuple[str, str | None, str | None]:
    """Detect the archive type and/or compression of a file.

    Args:
        file: the filename

    Returns:
        Tuple of suffix, archive type, and compression

    Raises:
        RuntimeError: if file has no suffix or suffix is not supported
    """
    suffixes = Path(file).suffixes
    if not suffixes:
        raise RuntimeError(
            f"File '{file}' has no suffixes that could be used to detect the archive type and compression."
        )
    suffix = suffixes[-1]

    # check if the suffix is a known alias
    if suffix in _FILE_TYPE_ALIASES:
        return (suffix, *_FILE_TYPE_ALIASES[suffix])

    # check if the suffix is an archive type
    if suffix in _ARCHIVE_EXTRACTORS:
        return suffix, suffix, None

    # check if the suffix is a compression
    if suffix in _COMPRESSED_FILE_OPENERS:
        # check for suffix hierarchy
        if len(suffixes) > 1:
            suffix2 = suffixes[-2]

            # check if the suffix2 is an archive type
            if suffix2 in _ARCHIVE_EXTRACTORS:
                return suffix2 + suffix, suffix2, suffix

        return suffix, None, suffix

    valid_suffixes = sorted(
        set(_FILE_TYPE_ALIASES) | set(_ARCHIVE_EXTRACTORS) | set(_COMPRESSED_FILE_OPENERS)
    )
    raise RuntimeError(
        f"Unknown compression or archive type: '{suffix}'.\nKnown suffixes are: '{valid_suffixes}'."
    )


def _decompress(
    from_path: Path | str, to_path: Path | str | None = None, remove_finished: bool = False
) -> str:
    """Decompress a file.

    The compression is automatically detected from the file name.

    Args:
        from_path: Path to the file to be decompressed.
        to_path: Path to the decompressed file. If omitted, `from_path` without compression extension is used.
        remove_finished: If `True`, remove the file after the extraction.

    Returns:
        Path to the decompressed file.
    """
    suffix, archive_type, compression = _detect_file_type(from_path)
    if not compression:
        raise RuntimeError(f"Couldn't detect a compression from suffix {suffix}.")

    if to_path is None:
        to_path = str(from_path).replace(suffix, archive_type if archive_type is not None else "")

    compressed_file_opener = _COMPRESSED_FILE_OPENERS[compression]

    with compressed_file_opener(from_path, "rb") as rfh, open(to_path, "wb") as wfh:
        wfh.write(rfh.read())

    if remove_finished:
        os.remove(from_path)

    return str(to_path)


def extract_archive(
    from_path: str | PathLike,
    extract_root: str | PathLike | None = None,
    members: list | None = None,
    remove_finished: bool = False,
) -> str:
    """Extract an archive.

    The archive type and a possible compression is automatically detected from
    the file name.

    If the file is compressed but not an archive, the call is dispatched to `_decompress` function.

    Args:
        from_path: Path to the file to be extracted.
        extract_root: Path to the directory the file will be extracted to.
            The given directory will be created if not exist.
            If omitted, the directory of the archive file is used.
        members: Optional selection of members to extract. If not specified,
            all members are extracted.
            Members must be a subset of the list returned by
            - `zipfile.ZipFile.namelist()` or a list of strings for zip file
            - `tarfile.TarFile.getmembers()` for tar file
        remove_finished: If `True`, remove the file after the extraction.

    Returns:
        Path to the directory the file was extracted to.
    """
    from_path = Path(from_path)

    if extract_root is None:
        extract_root = from_path.parent
    else:
        extract_root = Path(extract_root)

    # create the extract directory if not exist
    extract_root.mkdir(exist_ok=True)

    logger.info(f"Extracting {from_path} to {extract_root}")
    suffix, archive_type, compression = _detect_file_type(from_path)
    if not archive_type:
        return _decompress(
            from_path,
            extract_root / from_path.name.replace(suffix, ""),
            remove_finished=remove_finished,
        )

    extractor = _ARCHIVE_EXTRACTORS[archive_type]

    extractor(str(from_path), str(extract_root), members, compression)
    if remove_finished:
        from_path.unlink()

    return str(extract_root)


def download_and_extract_archive(
    url: str,
    download_root: str | PathLike,
    extract_root: str | Path | None = None,
    filename: str | None = None,
    md5: str | None = None,
    remove_finished: bool = False,
) -> None:
    """Download an archive file and then extract it.

    This method is a wrapper of [`download_url`][nplinker.utils.download_url] and
    [`extract_archive`][nplinker.utils.extract_archive] functions.

    Args:
        url: URL to download file from
        download_root: Path to the directory to place downloaded
            file in. If it doesn't exist, it will be created.
        extract_root: Path to the directory the file
            will be extracted to. The given directory will be created if not exist.
            If omitted, the `download_root` is used.
        filename: Name to save the downloaded file under.
            If None, use the basename of the URL
        md5: MD5 checksum of the download. If None, do not check
        remove_finished: If `True`, remove the downloaded file
             after the extraction. Defaults to False.
    """
    download_root = Path(download_root)
    if extract_root is None:
        extract_root = download_root
    else:
        extract_root = Path(extract_root)
    if not filename:
        filename = Path(url).name

    download_url(url, download_root, filename, md5)

    archive = download_root / filename
    extract_archive(archive, extract_root, remove_finished=remove_finished)


def transform_to_full_path(p: str | PathLike) -> Path:
    """Transform a path to a full path.

    The path is expanded (i.e. the `~` will be replaced with actual path) and converted to an
    absolute path (i.e. `.` or `..` will be replaced with actual path).

    Args:
        p: The path to transform.

    Returns:
        The transformed full path.
    """
    # Multiple calls to `Path` are used to ensure static typing compatibility.
    p = Path(p).expanduser()
    p = Path(p).resolve()
    return Path(p)



from enum import Enum
from enum import unique


@unique
class ScoringMethod(Enum):
    """Enum class for scoring methods."""

    METCALF = "metcalf"
    ROSETTA = "rosetta"
    NPLCLASS = "nplclass"

    @classmethod
    def has_value(cls, value: str) -> bool:
        """Check if the enum has a value."""
        return any(value == item.value for item in cls)



from __future__ import annotations
import json
import logging
from os import PathLike
from jsonschema import validate
from nplinker.schemas import USER_STRAINS_SCHEMA
from ..genomics.utils import extract_mappings_original_genome_id_resolved_genome_id
from ..genomics.utils import extract_mappings_resolved_genome_id_bgc_id
from ..genomics.utils import extract_mappings_strain_id_original_genome_id
from ..genomics.utils import get_mappings_strain_id_bgc_id
from ..metabolomics.utils import extract_mappings_ms_filename_spectrum_id
from ..metabolomics.utils import extract_mappings_strain_id_ms_filename
from ..metabolomics.utils import get_mappings_strain_id_spectrum_id
from .strain import Strain
from .strain_collection import StrainCollection


logger = logging.getLogger(__name__)


def load_user_strains(json_file: str | PathLike) -> set[Strain]:
    """Load user specified strains from a JSON file.

    The JSON file will be validated against the schema
    [USER_STRAINS_SCHEMA][nplinker.schemas.USER_STRAINS_SCHEMA]

    The content of the JSON file could be, for example:
    ```
    {"strain_ids": ["strain1", "strain2"]}
    ```

    Args:
        json_file: Path to the JSON file containing user specified strains.

    Returns:
        A set of user specified strains.
    """
    with open(json_file, "r") as f:
        json_data = json.load(f)

    # validate json data
    validate(instance=json_data, schema=USER_STRAINS_SCHEMA)

    strains = set()
    for strain_id in json_data["strain_ids"]:
        strains.add(Strain(strain_id))

    return strains


def podp_generate_strain_mappings(
    podp_project_json_file: str | PathLike,
    genome_status_json_file: str | PathLike,
    genome_bgc_mappings_file: str | PathLike,
    gnps_file_mappings_file: str | PathLike,
    output_json_file: str | PathLike,
) -> StrainCollection:
    """Generate strain mappings JSON file for PODP pipeline.

    To get the strain mappings, we need to combine the following mappings:

    - strain_id <-> original_genome_id <-> resolved_genome_id <-> bgc_id
    - strain_id <-> MS_filename <-> spectrum_id

    These mappings are extracted from the following files:

    - "strain_id <-> original_genome_id" is extracted from `podp_project_json_file`.
    - "original_genome_id <-> resolved_genome_id" is extracted from `genome_status_json_file`.
    - "resolved_genome_id <-> bgc_id" is extracted from `genome_bgc_mappings_file`.
    - "strain_id <-> MS_filename" is extracted from `podp_project_json_file`.
    - "MS_filename <-> spectrum_id" is extracted from `gnps_file_mappings_file`.

    Args:
        podp_project_json_file: The path to the PODP project
            JSON file.
        genome_status_json_file: The path to the genome status
            JSON file.
        genome_bgc_mappings_file: The path to the genome BGC
            mappings JSON file.
        gnps_file_mappings_file: The path to the GNPS file
            mappings file (csv or tsv).
        output_json_file: The path to the output JSON file.

    Returns:
        The strain mappings stored in a StrainCollection object.

    See Also:
        - `extract_mappings_strain_id_original_genome_id`: Extract mappings
            "strain_id <-> original_genome_id".
        - `extract_mappings_original_genome_id_resolved_genome_id`: Extract mappings
            "original_genome_id <-> resolved_genome_id".
        - `extract_mappings_resolved_genome_id_bgc_id`: Extract mappings
            "resolved_genome_id <-> bgc_id".
        - `get_mappings_strain_id_bgc_id`: Get mappings "strain_id <-> bgc_id".
        - `extract_mappings_strain_id_ms_filename`: Extract mappings
            "strain_id <-> MS_filename".
        - `extract_mappings_ms_filename_spectrum_id`: Extract mappings
            "MS_filename <-> spectrum_id".
        - `get_mappings_strain_id_spectrum_id`: Get mappings "strain_id <-> spectrum_id".
    """
    # Get mappings strain_id <-> original_genome_id <-> resolved_genome_id <-> bgc_id
    mappings_strain_id_bgc_id = get_mappings_strain_id_bgc_id(
        extract_mappings_strain_id_original_genome_id(podp_project_json_file),
        extract_mappings_original_genome_id_resolved_genome_id(genome_status_json_file),
        extract_mappings_resolved_genome_id_bgc_id(genome_bgc_mappings_file),
    )

    # Get mappings strain_id <-> MS_filename <-> spectrum_id
    mappings_strain_id_spectrum_id = get_mappings_strain_id_spectrum_id(
        extract_mappings_strain_id_ms_filename(podp_project_json_file),
        extract_mappings_ms_filename_spectrum_id(gnps_file_mappings_file),
    )

    # Get mappings strain_id <-> bgc_id / spectrum_id
    mappings = mappings_strain_id_bgc_id.copy()
    for strain_id, spectrum_ids in mappings_strain_id_spectrum_id.items():
        if strain_id in mappings:
            mappings[strain_id].update(spectrum_ids)
        else:
            mappings[strain_id] = spectrum_ids.copy()

    # Create StrainCollection
    sc = StrainCollection()
    for strain_id, bgc_ids in mappings.items():
        if not sc.has_name(strain_id):
            strain = Strain(strain_id)
            for bgc_id in bgc_ids:
                strain.add_alias(bgc_id)
            sc.add(strain)
        else:
            # strain_list has only one element
            strain_list = sc.lookup(strain_id)
            for bgc_id in bgc_ids:
                strain_list[0].add_alias(bgc_id)

    # Write strain mappings JSON file
    sc.to_json(output_json_file)
    logger.info("Generated strain mappings JSON file: %s", output_json_file)

    return sc



from __future__ import annotations
import logging
import os
import shutil
from pathlib import Path
from nplinker.utils import download_and_extract_archive
from nplinker.utils import list_dirs
from nplinker.utils import list_files


logger = logging.getLogger(__name__)

MIBIG_METADATA_URL = "https://dl.secondarymetabolites.org/mibig/mibig_json_{version}.tar.gz"

_MD5_MIBIG_METADATA = {
    "1.0": "035a14e94d2733eb61f615f418c08494",
    "1.1": "e63ceca82363ac27d50650e133ae3fa1",
    "1.2": "46e862018bd076d0b6072e62b7d8cfa2",
    "1.3": "94b3a761323709d06b663232f30210b4",
    "1.4": "a85530571d9dd7978b1bb0f2580cd30e",
    "2.0": "843ce4677db6d11422f0e6d94dd03e81",
    "3.0": "7c38b90f939086c03392d99a913baef9",
    "3.1": "643d1349722a9437d8dcf558dac5f815",
    "4.0": "70d1e7d573652ba62548b1fcfbdbf844",
}


def download_and_extract_mibig_metadata(
    download_root: str | os.PathLike,
    extract_path: str | os.PathLike,
    version: str = "3.1",
):
    """Download and extract MIBiG metadata json files.

    The MIBiG metadata json files are available at https://mibig.secondarymetabolites.org/download.

    Note that it does not matter whether the metadata json files are in nested folders or not in the archive,
    all json files will be extracted to the same location, i.e. `extract_path`. The nested
    folders will be removed if they exist. So the `extract_path` will have only json files.

    Args:
        download_root: Path to the directory in which to place the downloaded archive.
        extract_path: Path to an empty directory where the json files will be extracted.
            The directory must be empty if it exists. If it doesn't exist, the directory will be created.
        version: MIBiG version. Defaults to "3.1".

    Examples:
        >>> download_and_extract_mibig_metadata("/data/download", "/data/mibig_metadata")
    """
    download_root = Path(download_root)
    extract_path = Path(extract_path)

    if download_root == extract_path:
        raise ValueError("Identical path of download directory and extract directory")

    # check if extract_path is empty
    if not extract_path.exists():
        extract_path.mkdir(parents=True)
    else:
        if len(list(extract_path.iterdir())) != 0:
            raise ValueError(f'Nonempty directory: "{extract_path}"')

    # download and extract
    md5 = _MD5_MIBIG_METADATA.get(version, None)
    download_and_extract_archive(
        url=MIBIG_METADATA_URL.format(version=version),
        download_root=download_root,
        extract_root=extract_path,
        md5=md5,
    )

    # After extracting mibig archive, it's either one dir or many json files,
    # if it's a dir, then move all json files from it to extract_path
    subdirs = list_dirs(extract_path)
    if len(subdirs) > 1:
        raise ValueError(f"Expected one extracted directory, got {len(subdirs)}")

    if len(subdirs) == 1:
        subdir_path = subdirs[0]
        for fname in list_files(subdir_path, prefix="BGC", suffix=".json", keep_parent=False):
            shutil.move(os.path.join(subdir_path, fname), os.path.join(extract_path, fname))
        # delete subdir
        if subdir_path != extract_path:
            shutil.rmtree(subdir_path)



from .antismash_downloader import download_and_extract_antismash_data
from .antismash_loader import AntismashBGCLoader
from .antismash_loader import parse_bgc_genbank
from .podp_antismash_downloader import GenomeStatus
from .podp_antismash_downloader import get_best_available_genome_id
from .podp_antismash_downloader import podp_download_and_extract_antismash_data


__all__ = [
    "download_and_extract_antismash_data",
    "AntismashBGCLoader",
    "parse_bgc_genbank",
    "GenomeStatus",
    "get_best_available_genome_id",
    "podp_download_and_extract_antismash_data",
]



from __future__ import annotations
import logging
from os import PathLike
from pathlib import Path
from nplinker.strain import Strain
from nplinker.utils import list_files
from ..abc import BGCLoaderBase
from ..bgc import BGC
from .mibig_metadata import MibigMetadata


logger = logging.getLogger(__name__)


class MibigLoader(BGCLoaderBase):
    """Parse MIBiG metadata files and return BGC objects.

    MIBiG metadata file (json) contains annotations/metadata information
    for each BGC. See https://mibig.secondarymetabolites.org/download.

    The MiBIG accession is used as BGC id and strain name. The loaded BGC
    objects have Strain object as their strain attribute (i.e. `BGC.strain`).
    """

    def __init__(self, data_dir: str | PathLike):
        """Initialize the MIBiG metadata loader.

        Args:
            data_dir: Path to the directory of MIBiG metadata json files

        Examples:
            >>> loader = MibigLoader("path/to/mibig/data/dir")
            >>> loader.data_dir
            'path/to/mibig/data/dir'
            >>> loader.get_bgcs()
            [BGC('BGC000001', 'NRP'), BGC('BGC000002', 'Polyketide')]
        """
        self.data_dir = str(data_dir)
        self._file_dict = self.parse_data_dir(self.data_dir)
        self._metadata_dict = self._parse_metadata()
        self._bgcs = self._parse_bgcs()

    def get_files(self) -> dict[str, str]:
        """Get the path of all MIBiG metadata json files.

        Returns:
            The key is metadata file name (BGC accession), and the value is path to the metadata
            json file
        """
        return self._file_dict

    @staticmethod
    def parse_data_dir(data_dir: str | PathLike) -> dict[str, str]:
        """Parse metadata directory and return paths to all metadata json files.

        Args:
            data_dir: path to the directory of MIBiG metadata json files

        Returns:
            The key is metadata file name (BGC accession), and the value is path to the metadata
            json file
        """
        file_dict = {}
        json_files = list_files(data_dir, prefix="BGC", suffix=".json")
        for file in json_files:
            fname = Path(file).stem
            file_dict[fname] = file
        return file_dict

    def get_metadata(self) -> dict[str, MibigMetadata]:
        """Get MibigMetadata objects.

        Returns:
            The key is BGC accession (file name) and the value is MibigMetadata object
        """
        return self._metadata_dict

    def _parse_metadata(self) -> dict[str, MibigMetadata]:
        """Parse all metadata files and return MibigMetadata objects.

        Returns:
            The key is BGC accession (file name) and the value is MibigMetadata object
        """
        metadata_dict = {}
        for name, file in self._file_dict.items():
            metadata = MibigMetadata(file)
            metadata_dict[name] = metadata
        return metadata_dict

    def get_bgcs(self) -> list[BGC]:
        """Get BGC objects.

        The BGC objects use MiBIG accession as id and have Strain object as
        their strain attribute (i.e. `BGC.strain`), where the name of the Strain
        object is also MiBIG accession.

        Returns:
            A list of BGC objects
        """
        return self._bgcs

    def _parse_bgcs(self) -> list[BGC]:
        """Parse all metadata files as BGC objects.

        Returns:
            A list of BGC objects
        """
        return [parse_bgc_metadata_json(file) for file in self._file_dict.values()]


def parse_bgc_metadata_json(file: str | PathLike) -> BGC:
    """Parse MIBiG metadata file and return BGC object.

    Note that the MiBIG accession is used as the BGC id and strain name. The BGC
    object has Strain object as its strain attribute.

    Args:
        file: Path to the MIBiG metadata json file

    Returns:
        BGC object
    """
    metadata = MibigMetadata(str(file))
    mibig_bgc = BGC(metadata.mibig_accession, *metadata.biosyn_class)
    mibig_bgc.mibig_bgc_class = metadata.biosyn_class
    mibig_bgc.strain = Strain(metadata.mibig_accession)
    return mibig_bgc



# Copyright 2021 The NPLinker Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import csv
import logging
import os
from nplinker.defaults import NPLINKER_APP_DATA_DIR
from nplinker.scoring.rosetta.rosetta_hit import RosettaHit
from ...genomics import BGC
from ...parsers.kcb import KCBJSONParser
from ...parsers.kcb import KCBTextParser
from ...pickler import load_pickled_data
from ...pickler import save_pickled_data
from .spec_lib import SpecLib


logger = logging.getLogger(__name__)


class Rosetta:
    DEF_MS1_TOL = 100
    DEF_MS2_TOL = 0.2
    DEF_SCORE_THRESH = 0.5
    DEF_MIN_MATCH_PEAKS = 1

    PARAM_VERSION = 1

    def __init__(self, nplinker, ignore_genomic_cache=False):
        self._nplinker = nplinker
        self._mgf_data = {}
        self._csv_data = {}
        self._mgf_path = os.path.join(NPLINKER_APP_DATA_DIR, "matched_mibig_gnps_update.mgf")
        self._csv_path = os.path.join(NPLINKER_APP_DATA_DIR, "matched_mibig_gnps_update.csv")
        self._data_path = NPLINKER_APP_DATA_DIR
        self._root_path = nplinker.root_dir
        self._dataset_id = nplinker.dataset_id
        self._ignore_genomic_cache = ignore_genomic_cache
        self._pickle_dir = os.path.join(nplinker.root_dir, "rosetta")
        if not os.path.exists(self._pickle_dir):
            os.makedirs(self._pickle_dir, exist_ok=True)
        self._speclib_pickle_path = os.path.join(self._pickle_dir, "SpecLib.pckl")
        self._spechits_pickle_path = os.path.join(self._pickle_dir, "spec_hits.pckl")
        self._bgchits_pickle_path = os.path.join(self._pickle_dir, "bgc_hits.pckl")
        self._rhits_pickle_path = os.path.join(self._pickle_dir, "RosettaHits.pckl")
        self._params_pickle_path = os.path.join(self._pickle_dir, "RosettaParams.pckl")

        if not os.path.exists(self._mgf_path):
            logger.warning(
                "Failed to load Rosetta data ({}), matching disabled".format(self._mgf_path)
            )
            return
        if not os.path.exists(self._csv_path):
            logger.warning(
                "Failed to load Rosetta data ({}), matching disabled".format(self._csv_path)
            )
            return

        self._gnps2mibig = None
        self._mibig2gnps = None
        self._mibig2bgc = {}

        self.speclib = None
        self._rosetta_hits = []
        self._spec_hits = None
        self._bgc_hits = None

    @property
    def bgc_hits(self):
        return self._bgc_hits

    @property
    def spec_hits(self):
        return self._spec_hits

    def _load_csv(self, csv_path):
        logger.info("constructing rosetta dicts")

        self._gnps2mibig = {}
        self._mibig2gnps = {}

        with open(csv_path) as f:
            rdr = csv.reader(f, delimiter=",")
            headers = next(rdr)
            for line in rdr:
                gnps, mibig = line[0], line[3]
                if gnps in self._gnps2mibig:
                    self._gnps2mibig[gnps].append(mibig)
                else:
                    self._gnps2mibig[gnps] = [mibig]

                if mibig in self._mibig2gnps:
                    self._mibig2gnps[mibig].append(gnps)
                else:
                    self._mibig2gnps[mibig] = [gnps]

    def _generate_spec_hits(self, spectra, ms1_tol, ms2_tol, score_thresh, min_match_peaks):
        spec_hits = {}
        for i, sp in enumerate(spectra):
            hits = self.speclib.spectral_match(
                sp,
                ms2_tol=ms2_tol,
                min_match_peaks=min_match_peaks,
                ms1_tol=ms1_tol,
                score_thresh=score_thresh,
            )
            if len(hits) > 0:
                spec_hits[sp] = hits
            if i % 100 == 0:
                logger.info("Searching for spectral hits {}/{}".format(i, len(spectra)))

        save_pickled_data(spec_hits, self._spechits_pickle_path)
        return spec_hits

    def _generate_speclib(self):
        logger.warning("No pickle SpecLib found, generating (this will take some time!)...")
        self.speclib = SpecLib(self._mgf_path)
        self.speclib._load_mgf()
        self.speclib.filter()

        logger.info("Finished generating SpecLib")

        save_pickled_data(self.speclib, self._speclib_pickle_path)

    def _generate_bgc_hits(self, bgcs):
        self._bgc_hits = {}
        errors = 0

        # this method is a bit messy because it tries to handle a couple of different
        # routes to extracting knownclusterblast results:
        #
        #   - newer antiSMASH runs should produce a single .JSON file in each directory
        #       of .gbk files. this one file can be parsed to extract all the info that the
        #       rosetta code requires for all of the .gbks (uses KCBJSONParser)
        #   - older datasets will only include the now-deprecated text format results, which
        #       should be located in a subdirectory of each of the .gbk directories called
        #       "knownclusterblast". in these instances there will be a single .txt file
        #       for each .gbk
        #   - some datasets may have no knownclusterblast results available
        #
        # it's also possible for all 3 of these to turn up within the same dataset. for example
        # when downloading from the paired omics platform, the antiSMASH data is downloaded
        # separately for each genome from the antiSMASH DB and this means the format may not
        # be consistent.

        logger.debug("Collecting BGC hit information...")
        # go through the list of all available BGCs (ignoring MiBIGBGC instances) and
        # group them by the directory they appear in
        bgc_groups = {}
        skipped, errors = 0, 0

        for bgc in bgcs:
            if bgc.antismash_file is None or isinstance(bgc, BGC):
                skipped += 1
                continue

            prefix = os.path.dirname(bgc.antismash_file)
            if prefix not in bgc_groups:
                bgc_groups[prefix] = [bgc]
            else:
                bgc_groups[prefix].append(bgc)

        logger.debug("{} BGC groups based on filenames".format(len(bgc_groups)))

        for prefix, prefix_bgcs in bgc_groups.items():
            logger.debug(
                "Attempting to parse JSON data for prefix {} with {} BGCs".format(
                    prefix, len(prefix_bgcs)
                )
            )
            # preferred option is to parse the results for the whole group using the
            # JSON file, but this may not be available...
            json_hits = KCBJSONParser(prefix_bgcs).parse_hits()
            matched_bgcs = {}

            if json_hits is not None:
                # number of hits can often be less than number of BGCs (e.g. if no significant hits found)
                sum_hits = sum(len(json_hits[x]) for x in json_hits)
                logger.debug(
                    "JSON parsing was successful! Returned {} hits from {} BGCs".format(
                        sum_hits, len(prefix_bgcs)
                    )
                )

                # unlike the KCBTextParser where each set of results is easy to link
                # back to the appropriate BGC object, here we need to do some extra work
                # to ensure we have everything matched up correctly.
                #
                # seems like the best way to do this is to rely on the BGC attributes
                # parsed from the .gbks during the loading process (including region
                # numbers) as these should match up directly with the JSON data.

                for pbgc in prefix_bgcs:
                    # the "normal" case appears to be that you'll have a directory
                    # containing multiple gbks with the same accession and different
                    # region numbers, e.g. ABC123.region001, ABC123.region002, ...
                    # and should be able to expect that every "hit" comes from a
                    # .gbk that exists in the directory.
                    #
                    # in ideal circumstances, the json_hits structure will end up
                    # containing a single top level accession (ABC123) and then
                    # at the next level down one region number for each of the
                    # BGCs with signficant hits. this makes matching BGC objects
                    # quite simple.
                    #
                    # however in other cases there appear to be a mix of accession
                    # IDs in the same antiSMASH directory. so you can have collections
                    # where the filenames are e.g. ABC123.region001, DEF456.region001,
                    # GHI789.region001, ... (including multiple regions for the same
                    # accession). this is more difficult to handle because the JSON
                    # data isn't a direct match with that parsed from the .gbks
                    # themselves. for example, the gbk might report a region number of
                    # 7 while the corresponding JSON result has a region number of 1.
                    # possible workaround is to take the gbk region number and check
                    # if it appears in the filename of the gbk???

                    if pbgc.antismash_id in json_hits:
                        # simplest case where there's a direct match on region number
                        if pbgc.antismash_region in json_hits[pbgc.antismash_id]:
                            logger.debug(
                                "Matched {} using {} + region{:03d}!".format(
                                    pbgc.antismash_file, pbgc.antismash_id, pbgc.antismash_region
                                )
                            )
                            if pbgc not in matched_bgcs:
                                matched_bgcs[pbgc] = {}

                            hit = json_hits[pbgc.antismash_id][pbgc.antismash_region]
                            matched_bgcs[pbgc][hit["mibig_id"]] = hit
                            continue
                        else:
                            # if the above case doesn't apply, check through every
                            # region number available for the antismash ID we have,
                            # and check if the original filename contains that region
                            # number. if so assume it is the correct match.
                            for region in json_hits[pbgc.antismash_id]:
                                if pbgc.antismash_file.endswith(
                                    "region{:03d}.gbk".format(pbgc.antismash_region)
                                ):
                                    logger.debug(
                                        "Matched {} using fallback {} + region{:03d} (orig={})".format(
                                            pbgc.antismash_file,
                                            pbgc.antismash_id,
                                            region,
                                            pbgc.antismash_region,
                                        )
                                    )
                                    if pbgc not in matched_bgcs:
                                        matched_bgcs[pbgc] = {}

                                    hit = json_hits[pbgc.antismash_id][region]
                                    matched_bgcs[pbgc][hit["mibig_id"]] = hit
                                    break
                    else:
                        # this could simply mean no significant hits found
                        logger.info(
                            "Found no matching hits for BGC ID={}, region={}, file={}".format(
                                pbgc.antismash_id, pbgc.antismash_region, pbgc.antismash_file
                            )
                        )

            else:
                # ... if JSON parsing failed, fall back to the original text parser. this
                # must be called on each BGC individually
                logger.debug("JSON parsing failed, falling back to text instead")
                for i, bgc in enumerate(prefix_bgcs):
                    kcb_name = KCBTextParser.get_kcb_filename_from_bgc(bgc)
                    if kcb_name is not None:
                        try:
                            parser = KCBTextParser(kcb_name)
                            if len(parser.hits) > 0:
                                matched_bgcs[bgc] = parser.hits
                        except Exception as e:
                            logger.warning(e)
                            errors += 1

            logger.debug("Found matches for {}/{} bgcs".format(len(matched_bgcs), len(prefix_bgcs)))
            if len(matched_bgcs) != len(prefix_bgcs):
                # not necessarily fatal but probably not good either
                logger.warning(
                    "Failed to match {} BGCs to hits in directory {}!".format(
                        len(prefix_bgcs) - len(matched_bgcs), prefix
                    )
                )

            # now insert the matched hits into the _bgc_hits structure so the original code
            # below can parse them in the same way as the text parser results
            self._bgc_hits.update(matched_bgcs)

        # make reverse dict
        self._mibig2bgc = {}
        for bgc, hits in self._bgc_hits.items():
            for mibig_bgc_id in hits:
                if mibig_bgc_id not in self._mibig2bgc:
                    self._mibig2bgc[mibig_bgc_id] = set()
                self._mibig2bgc[mibig_bgc_id].add(bgc)

        logger.info(f"Completed, {len(self._bgc_hits)} BGC hits found")
        if errors > 0:
            logger.warning(
                "Some knownclusterblast files could not be loaded, results may be incomplete"
            )
        save_pickled_data((self._bgc_hits, self._mibig2bgc), self._bgchits_pickle_path)

    def _collect_rosetta_hits(self):
        self._rosetta_hits = []
        bgc_summary_scores = self.generate_bgc_summary_scores()
        for spec, data in self._spec_hits.items():
            for gnps_id, score in data:
                for mibig_id in self._gnps2mibig[gnps_id]:
                    if mibig_id in self._mibig2bgc:
                        for bgc in self._mibig2bgc[mibig_id]:
                            # get the bgc score
                            bgc_score = bgc_summary_scores[bgc][mibig_id]
                            self._rosetta_hits.append(
                                RosettaHit(spec, gnps_id, mibig_id, bgc, score, bgc_score)
                            )
        logger.info(f"Found {len(self._rosetta_hits)} rosetta hits!")

    def generate_bgc_summary_scores(self):
        # process the hit to compress it into more useful info
        # computes the total identity score for each mibig entry
        # and divides by the number of mibig genes
        # i.e. the score represents how much of the mibig is
        # reflected in the source bgc

        # we have one entry per mibig that is linked to this BGC and
        # a single score (that will vary between 0 and 1)

        processed = {}
        # bgc_hits = {BGC object: list of hits from KCB}
        # each list is a list of dicts
        for bgc, hit in self._bgc_hits.items():
            mibig_bgcs = list(hit.keys())
            scores = {}
            for mibig_id in mibig_bgcs:
                n_mibig_genes = len(hit[mibig_id]["all_mibig_genes"])
                if n_mibig_genes == 0:
                    logger.warning(
                        "Found a BGC entry with zero genes, this should never happen! (BGC={})".format(
                            bgc
                        )
                    )
                    continue
                # n_source_genes = len(hit[mibig_id]['all_bgc_genes'])
                total_hit_identity = 0
                for hit_gene in hit[mibig_id]["individual_hits"]:
                    identity_percent = hit_gene["identity_percent"]
                    total_hit_identity += identity_percent / 100.0
                score = total_hit_identity / n_mibig_genes
                scores[mibig_id] = score
            processed[bgc] = scores
        return processed

    def run(self, spectra, bgcs, ms1_tol, ms2_tol, score_thresh, min_match_peaks):
        """Function which actually computes the rosetta score somehow."""
        params_ok = self._load_cached_params(ms1_tol, ms2_tol, score_thresh, min_match_peaks)

        # if any parameters have been changed or version mismatch found, delete all cached files
        if not params_ok:
            self._clear_cache(ms1_tol, ms2_tol, score_thresh, min_match_peaks)

            self.speclib = None
            self._spec_hits = None
            self._bgc_hits = None
            self._rosetta_hits = []

        # next, try to load the cached rosetta_hits list. if parameters were invalidated above,
        # the file will have been deleted and this will fail
        logger.info("Trying to load cached Rosetta hits data")
        cached_rosetta_hits = load_pickled_data(self._nplinker, self._rhits_pickle_path)
        if cached_rosetta_hits is not None:
            logger.info(
                "Loaded cached Rosetta hits for dataset {} at {}".format(
                    self._dataset_id, self._rhits_pickle_path
                )
            )
            self._rosetta_hits = cached_rosetta_hits
            return self._rosetta_hits

        # if we get this far, it means regenerating some/all of the required data structures.
        #
        # create the _gnps2mibig and _mibig2gnps dicts if not already done
        if self._mibig2gnps is None or self._gnps2mibig is None:
            logger.info("Constructing GNPS/MiBIG dicts")
            self._load_csv(self._csv_path)

        self._init_bgc_hits(bgcs)

        # if we didn't find any BGC hits, no point in continuing
        if len(self._bgc_hits) == 0:
            logger.warning("Aborting Rosetta scoring data generation, no BGC hits were found!")
            # create an empty rosetta_hits.csv file
            self.export_to_csv(os.path.join(self._pickle_dir, "rosetta_hits.csv"))
            return self._rosetta_hits

        logger.info("No cached Rosetta hits data found")

        self._init_speclib(spectra, ms1_tol, ms2_tol, score_thresh, min_match_peaks)

        self._init_spec_hits(spectra, ms1_tol, ms2_tol, score_thresh, min_match_peaks)

        # finally construct the list of rosetta hits
        self._collect_rosetta_hits()

        # export cached data for future runs
        save_pickled_data(self._rosetta_hits, self._rhits_pickle_path)
        save_pickled_data(
            (Rosetta.PARAM_VERSION, ms1_tol, ms2_tol, score_thresh, min_match_peaks),
            self._params_pickle_path,
        )

        # automatically export CSV file containing hit data to <dataset>/rosetta
        # along with the pickled data
        self.export_to_csv(os.path.join(self._pickle_dir, "rosetta_hits.csv"))

        return self._rosetta_hits

    def _init_bgc_hits(self, bgcs):
        """Collect BGC hits. this is done first because the SpecLib generation below can take
        several minutes and is a waste of time if the knownclusterblast files required for
        the genomics data aren't available in the current dataset.
        """
        cached_bgc_hits = load_pickled_data(self._nplinker, self._bgchits_pickle_path)
        if cached_bgc_hits is not None and not self._ignore_genomic_cache:
            logger.info("Found pickled bgc_hits for dataset {}!".format(self._dataset_id))
            self._bgc_hits, self._mibig2bgc = cached_bgc_hits
        else:
            logger.info("Generating BGC hits")
            self._generate_bgc_hits(bgcs)

    def _init_spec_hits(self, spectra, ms1_tol, ms2_tol, score_thresh, min_match_peaks):
        spec_hits = load_pickled_data(self._nplinker, self._spechits_pickle_path)
        if spec_hits is not None:
            logger.info(
                "Found pickled spectral hits for dataset {} at {}".format(
                    self._dataset_id, self._spechits_pickle_path
                )
            )
            self._spec_hits = spec_hits

        if self._spec_hits is None:
            # no cached spectral hits available, generate (and cache)
            logger.info("Generating spectral hits")
            self._spec_hits = self._generate_spec_hits(
                spectra, ms1_tol, ms2_tol, score_thresh, min_match_peaks
            )

        logger.info(
            "SpecLib has {} spectra, {} hits".format(
                self.speclib.get_n_spec(), len(self._spec_hits)
            )
        )

    def _init_speclib(self, spectra, ms1_tol, ms2_tol, score_thresh, min_match_peaks):
        """Next is the metabolomic part. check if we have a pickled SpecLib object..."""
        speclib = load_pickled_data(self._nplinker, self._speclib_pickle_path)
        if speclib is not None:
            logger.info(
                "Found pickled SpecLib for dataset {} at {}!".format(
                    self._dataset_id, self._speclib_pickle_path
                )
            )
            self.speclib = speclib

        if self.speclib is None:
            # no cached speclib available, generate (and cache)
            logger.info("Generating SpecLib")
            self._generate_speclib(spectra)

    def _load_cached_params(self, ms1_tol, ms2_tol, score_thresh, min_match_peaks):
        """Check if cached parameters exist, and if so check they match the
        supplied ones. if not, need to regenerate any pickled data files.
        """
        params = load_pickled_data(self._nplinker, self._params_pickle_path)
        params_ok = False

        if params is not None:
            try:
                if params[0] != Rosetta.PARAM_VERSION:
                    logger.warning(
                        "Rosetta: pickled data version mismatch (old {}, new {})".format(
                            params[0], Rosetta.PARAM_VERSION
                        )
                    )
                else:
                    _version, _ms1_tol, _ms2_tol, _score_thresh, _min_match_peaks = params

                    if (
                        ms1_tol == _ms1_tol
                        and ms2_tol == _ms2_tol
                        and score_thresh == _score_thresh
                        and min_match_peaks == _min_match_peaks
                    ):
                        # params only valid if all of these match up
                        params_ok = True

            except Exception as e:
                logger.warning(f"Failed to parse pickled Rosetta parameters: {e}")

        return params_ok

    def _clear_cache(self, ms1_tol, ms2_tol, score_thresh, min_match_peaks):
        logger.info(
            "SpecLib parameters have been changed or do not exist, regenerating cached data files!"
        )
        logger.debug(
            "ms1_tol={:.3f}, ms2_tol={:.3f}, score_thresh={:.3f}, min_match_peaks={:d}".format(
                ms1_tol, ms2_tol, score_thresh, min_match_peaks
            )
        )
        for path in [
            self._bgchits_pickle_path,
            self._spechits_pickle_path,
            self._rhits_pickle_path,
            self._params_pickle_path,
            self._speclib_pickle_path,
            os.path.join(self._pickle_dir, "rosetta_hits.csv"),
        ]:
            if os.path.exists(path):
                os.unlink(path)

    def export_to_csv(self, filename):
        # convenience method for exporting a full set of rosetta hits to a CSV file
        with open(filename, "w", newline="") as csvfile:
            csvwriter = csv.writer(csvfile, delimiter=",")

            csvwriter.writerow(
                [
                    "nplinker spectrum ID",
                    "spectrum ID",
                    "GNPS ID",
                    "spectral score",
                    "BGC ID",
                    "MiBIG BGC ID",
                    "BGC score",
                ]
            )
            for hit in self._rosetta_hits:
                csvwriter.writerow(
                    [
                        hit.spec.id,
                        hit.gnps_id,
                        hit.spec_match_score,
                        hit.bgc.id,
                        hit.mibig_id,
                        hit.bgc_match_score,
                    ]
                )



from __future__ import annotations
import logging
import os
import shutil
from os import PathLike
from pathlib import Path
from nplinker.utils import download_and_extract_archive
from nplinker.utils import list_dirs
from nplinker.utils import list_files


logger = logging.getLogger(__name__)

# urls to be given to download antismash data
ANTISMASH_DB_DOWNLOAD_URL = "https://antismash-db.secondarymetabolites.org/output/{}/{}"
# The antiSMASH DBV2 is for the availability of the old version, better to keep it.
ANTISMASH_DBV2_DOWNLOAD_URL = "https://antismash-dbv2.secondarymetabolites.org/output/{}/{}"


def download_and_extract_antismash_data(
    antismash_id: str, download_root: str | PathLike, extract_root: str | PathLike
) -> None:
    """Download and extract antiSMASH BGC archive for a specified genome.

    The antiSMASH database (https://antismash-db.secondarymetabolites.org/)
    is used to download the BGC archive. And antiSMASH use RefSeq assembly id
    of a genome as the id of the archive.

    Args:
        antismash_id: The id used to download BGC archive from antiSMASH database.
            If the id is versioned (e.g., "GCF_004339725.1") please be sure to
            specify the version as well.
        download_root: Path to the directory to place downloaded archive in.
        extract_root: Path to the directory data files will be extracted to.
            Note that an `antismash` directory will be created in the specified `extract_root` if
            it doesn't exist. The files will be extracted to `<extract_root>/antismash/<antismash_id>` directory.

    Raises:
        ValueError: if `<extract_root>/antismash/<refseq_assembly_id>` dir is not empty.

    Examples:
        >>> download_and_extract_antismash_metadata("GCF_004339725.1", "/data/download", "/data/extracted")
    """
    download_root = Path(download_root)
    extract_root = Path(extract_root)
    extract_path = extract_root / "antismash" / antismash_id

    try:
        if extract_path.exists():
            _check_extract_path(extract_path)
        else:
            extract_path.mkdir(parents=True, exist_ok=True)

        for base_url in [ANTISMASH_DB_DOWNLOAD_URL, ANTISMASH_DBV2_DOWNLOAD_URL]:
            url = base_url.format(antismash_id, antismash_id + ".zip")
            download_and_extract_archive(url, download_root, extract_path, antismash_id + ".zip")
            break

        # delete subdirs
        for subdir_path in list_dirs(extract_path):
            shutil.rmtree(subdir_path)

        # delete unnecessary files
        files_to_keep = list_files(extract_path, suffix=(".json", ".gbk"))
        for file in list_files(extract_path):
            if file not in files_to_keep:
                os.remove(file)

        logger.info("antiSMASH BGC data of %s is downloaded and extracted.", antismash_id)

    except Exception as e:
        shutil.rmtree(extract_path)
        logger.warning(e)
        raise e


def _check_extract_path(extract_path: Path):
    # check if extract_path is empty
    if any(extract_path.iterdir()):
        raise ValueError(f'Nonempty directory: "{extract_path}"')



# Copyright 2021 The NPLinker Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import multiprocessing
import os
import time
import iokrdata as data
import numpy
import scipy
import spectrum_filters
from spectrum import MSSpectrum
from spectrum import ppk
from spectrum import ppk_nloss


def ppk_r(spec1, spec2, prec1, prec2, sigma_mass, sigma_int):
    k_peaks = ppk(spec1, spec2, sigma_mass, sigma_int)
    k_nloss = ppk_nloss(spec1, spec2, prec1, prec2, sigma_mass, sigma_int)
    k_diff = ppk_diff(spec1, spec2, sigma_mass, sigma_int)
    return k_peaks + k_nloss + k_diff


def ppk_diff(spec1, spec2, sigma_mass, sigma_int):
    spec1_diff = numpy.array([y - x for x, y in itertools.combinations(spec1, 2)])
    spec2_diff = numpy.array([y - x for x, y in itertools.combinations(spec2, 2)])
    k_diff = ppk(spec1_diff, spec2_diff, sigma_mass, sigma_int)
    return k_diff


def create_ppk_matrix():
    iokr_data_path = "/home/grimur/iokr/data"
    data_gnps = scipy.io.loadmat("/home/grimur/iokr/data/data_GNPS.mat")
    ms_path = "/home/grimur/iokr/data/SPEC"

    sigma_mass = 0.00001
    sigma_int = 100000.0

    iokrdata = data.IOKRDataServer(iokr_data_path, kernel="PPKr.txt")
    ker_size = len(iokrdata.spectra)

    kernel_matrix_peaks = numpy.zeros((ker_size, ker_size))
    kernel_matrix_nloss = numpy.zeros((ker_size, ker_size))
    kernel_matrix_ppkr = numpy.zeros((ker_size, ker_size))

    j_ms = MSSpectrum()
    j_ms.filter = spectrum_filters.filter_by_collected_dag
    i_ms = MSSpectrum()
    i_ms.filter = spectrum_filters.filter_by_collected_dag

    import time

    t0 = time.time()
    cnt = 0
    for i in range(len(iokrdata.spectra)):
        i_name = iokrdata.spectra[i][0]
        i_ms.load(ms_path + os.sep + i_name + ".ms")
        # for j in range(i + 1):
        if True:
            j = i
            cnt += 1

            j_name = iokrdata.spectra[j][0]
            j_ms.load(ms_path + os.sep + j_name + ".ms")

            # print('%s vs %s' % (len(i_ms.spectrum), len(j_ms.spectrum)))

            if len(i_ms.spectrum) == 0 or len(j_ms.spectrum) == 0:
                print("empty")
                ij_peaks = 0
                ij_nloss = 0
            else:
                ij_peaks = ppk(i_ms.spectrum, j_ms.spectrum, sigma_mass, sigma_int)
                ij_nloss = ppk_nloss(
                    i_ms.spectrum,
                    j_ms.spectrum,
                    i_ms.parentmass,
                    j_ms.parentmass,
                    sigma_mass,
                    sigma_int,
                )

            kernel_matrix_peaks[i, j] = ij_peaks
            kernel_matrix_peaks[j, i] = ij_peaks

            kernel_matrix_nloss[i, j] = ij_nloss
            kernel_matrix_nloss[j, i] = ij_nloss

            kernel_matrix_ppkr[i, j] = ij_peaks + ij_nloss
            kernel_matrix_ppkr[j, i] = ij_peaks + ij_nloss

            if cnt % 100 == 0:
                print("done {}/{}, {}".format(cnt, (ker_size**2) / 2, time.time() - t0))
                t0 = time.time()

    # numpy.savetxt('ppk_peaks.csv', kernel_matrix_peaks, delimiter=',')
    # numpy.savetxt('ppk_nloss.csv', kernel_matrix_nloss, delimiter=',')
    # numpy.savetxt('ppk_r.csv', kernel_matrix_ppkr, delimiter=',')
    numpy.save("ppk_dag_peaks.npy", kernel_matrix_peaks)
    numpy.save("ppk_dag_nloss.npy", kernel_matrix_nloss)


def create_ppk_matrix_stripe_serial(filter_func, shift, normalise, output_name):
    iokr_data_path = "/home/grimur/iokr/data"
    data_gnps = scipy.io.loadmat("/home/grimur/iokr/data/data_GNPS.mat")
    ms_path = "/home/grimur/iokr/data/SPEC"

    iokrdata = data.IOKRDataServer(iokr_data_path)
    ker_size = len(iokrdata.spectra)

    kernel_matrix_peaks = numpy.zeros((ker_size, ker_size))
    kernel_matrix_nloss = numpy.zeros_like(kernel_matrix_peaks)

    p = multiprocessing.Pool(8)
    active_jobs = []

    t0 = time.time()
    names = [x[0] for x in iokrdata.spectra]
    cnt = 0
    for i in range(len(iokrdata.spectra)):
        if i == len(iokrdata.spectra) - 1:
            wait_for_clear = 0
        else:
            wait_for_clear = 10

        # active_jobs.append((i, p.apply_async(do_stripe, (i, names))))
        res = do_stripe(i, names, filter_func, shift, normalise)
        i_res = i

        if True:
            # if len(active_jobs) > wait_for_clear:
            #     active_jobs, results = gather_results_2(active_jobs, queue_length=wait_for_clear)
            #     for i_res, res in results:
            for j_res, values in enumerate(res):
                ij_peaks, ij_nloss = values

                kernel_matrix_peaks[i_res, j_res] = ij_peaks
                kernel_matrix_peaks[j_res, i_res] = ij_peaks

                kernel_matrix_nloss[i_res, j_res] = ij_nloss
                kernel_matrix_nloss[j_res, i_res] = ij_nloss

                cnt += 1
                if cnt % 100 == 0:
                    print("done %s/%s, %s" % (cnt, (ker_size**2) / 2, time.time() - t0))
                    t0 = time.time()

    numpy.save(output_name + "_peaks.npy", kernel_matrix_peaks)
    numpy.save(output_name + "_nloss.npy", kernel_matrix_nloss)


def gather_results(active_jobs):
    while len(active_jobs) > 500:
        done_jobs = []
        remaining_jobs = []
        for i, j, job in active_jobs:
            if job.ready():
                res = job.get()
                done_jobs.append((i, j, res))
            else:
                remaining_jobs.append((i, j, job))
        active_jobs = remaining_jobs
        # time.sleep(1)
    return active_jobs, done_jobs


def gather_results_2(active_jobs, queue_length):
    while len(active_jobs) > queue_length:
        done_jobs = []
        remaining_jobs = []
        for i, job in active_jobs:
            if job.ready():
                res = job.get()
                done_jobs.append((i, res))
            else:
                remaining_jobs.append((i, job))
        active_jobs = remaining_jobs
        time.sleep(0.5)
    return active_jobs, done_jobs


def do_stripe(i, names, filter_func, shift, normalise):
    iokr_data_path = "/home/grimur/iokr/data"
    data_gnps = scipy.io.loadmat("/home/grimur/iokr/data/data_GNPS.mat")
    ms_path = "/home/grimur/iokr/data/SPEC"

    sigma_mass = 0.00001
    sigma_int = 100000.0

    i_ms = MSSpectrum()
    i_ms.correct_for_ionisation = shift
    i_ms.filter = filter_func
    i_ms.normalise = normalise
    j_ms = MSSpectrum()
    j_ms.correct_for_ionisation = shift
    j_ms.filter = filter_func
    j_ms.normalise = normalise

    i_ms.load(ms_path + os.sep + names[i] + ".ms")
    results = []
    for j in range(i + 1):
        j_ms.load(ms_path + os.sep + names[j] + ".ms")

        ij_peaks = ppk(i_ms.spectrum, j_ms.spectrum, sigma_mass, sigma_int)
        ij_nloss = ppk_nloss(
            i_ms.spectrum, j_ms.spectrum, i_ms.parentmass, j_ms.parentmass, sigma_mass, sigma_int
        )

        results.append((ij_peaks, ij_nloss))

    return results


def do_pair(i_spectrum, j_spectrum, i_parentmass, j_parentmass, sigma_mass, sigma_int):
    if len(i_spectrum) == 0 or len(j_spectrum) == 0:
        print("empty spectrum!")
        return 0, 0
    ij_peaks = ppk(i_spectrum, j_spectrum, sigma_mass, sigma_int)
    ij_nloss = ppk_nloss(i_spectrum, j_spectrum, i_parentmass, j_parentmass, sigma_mass, sigma_int)
    # ij_diff = ppk_diff(i_spectrum, j_spectrum, sigma_mass, sigma_int)
    return ij_peaks, ij_nloss  # , ij_diff


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("-f", dest="filter_type", help="filter type (dag, tree)", default=None)
    parser.add_argument(
        "-c",
        dest="collected",
        help="build filter from all input files",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "-s", dest="shift", help="correct for ionisation", action="store_true", default=False
    )
    parser.add_argument("-o", dest="output", help="output name", required=True)
    parser.add_argument(
        "-n", dest="normalise", help="normalise", action="store_true", default=False
    )
    args = parser.parse_args()

    if args.filter_type == "dag":
        if args.collected:
            filter_func = spectrum_filters.filter_by_collected_dag
        else:
            filter_func = spectrum_filters.filter_by_dag
    elif args.filter_type == "tree":
        if args.collected:
            filter_func = spectrum_filters.filter_by_collected_tree
        else:
            if args.shift == False:
                filter_func = spectrum_filters.filter_by_tree_unshifted
            else:
                filter_func = spectrum_filters.filter_by_tree
    elif args.filter_type is None:
        filter_func = None
    else:
        raise SystemExit("Unknown filter: %s" % args.filter_type)

    # create_ppk_matrix_parallell()
    create_ppk_matrix_stripe_serial(filter_func, args.shift, args.normalise, args.output)
    # create_ppk_matrix()



import logging
from .logger import setup_logging
from .nplinker import NPLinker


logging.getLogger(__name__).addHandler(logging.NullHandler())

__author__ = "Cunliang Geng"
__email__ = "c.geng@esciencecenter.nl"
__version__ = "2.0.0-alpha.7"


__all__ = ["NPLinker", "setup_logging"]



from __future__ import annotations
import logging
from abc import ABC
from abc import abstractmethod
from typing import TYPE_CHECKING
from .link_graph import LinkGraph


if TYPE_CHECKING:
    from nplinker.nplinker import NPLinker

logger = logging.getLogger(__name__)


class ScoringBase(ABC):
    """Abstract base class of scoring methods.

    Attributes:
        name: The name of the scoring method.
        npl: The NPLinker object.
    """

    name: str = "ScoringBase"
    npl: NPLinker | None = None

    @classmethod
    @abstractmethod
    def setup(cls, npl: NPLinker):
        """Setup class level attributes."""

    @abstractmethod
    def get_links(
        self,
        *objects,
        **parameters,
    ) -> LinkGraph:
        """Get links information for the given objects.

        Args:
            objects: A list of objects to get links for.
            parameters: The parameters used for scoring.

        Returns:
            The LinkGraph object.
        """

    @abstractmethod
    def format_data(self, data) -> str:
        """Format the scoring data to a string."""

    @abstractmethod
    def sort(self, objects, reverse=True) -> list:
        """Sort the given objects based on the scoring data."""



from __future__ import annotations
import logging


logger = logging.getLogger(__name__)


class Strain:
    """Class to model the mapping between strain id and its aliases.

    It's recommended to use NCBI taxonomy strain id or name as the primary
    id.


    Attributes:
        id: The representative id of the strain.
        names: A set of names associated with the strain.
        aliases: A set of aliases associated with the strain.
    """

    def __init__(self, id: str) -> None:
        """To model the mapping between strain id and its aliases.

        Args:
            id: the representative id of the strain.
        """
        self.id: str = id
        self._aliases: set[str] = set()

    def __repr__(self) -> str:
        return str(self)

    def __str__(self) -> str:
        return f"Strain({self.id}) [{len(self._aliases)} aliases]"

    def __eq__(self, other) -> bool:
        if isinstance(other, Strain):
            return self.id == other.id
        return NotImplemented

    def __hash__(self) -> int:
        """Hash function for Strain.

        Note that Strain is a mutable container, so here we hash on only the id
        to avoid the hash value changes when `self._aliases` is updated.
        """
        return hash(self.id)

    def __contains__(self, alias: str) -> bool:
        if not isinstance(alias, str):
            raise TypeError(f"Expected str, got {type(alias)}")
        return alias in self._aliases

    @property
    def names(self) -> set[str]:
        """Get the set of strain names including id and aliases.

        Returns:
            A set of names associated with the strain.
        """
        return self._aliases | {self.id}

    @property
    def aliases(self) -> set[str]:
        """Get the set of known aliases.

        Returns:
            A set of aliases associated with the strain.
        """
        return self._aliases

    def add_alias(self, alias: str) -> None:
        """Add an alias for the strain.

        Args:
            alias: The alias to add for the strain.
        """
        if not isinstance(alias, str):
            raise TypeError(f"Expected str, got {type(alias)}")
        if len(alias) == 0:
            logger.warning("Refusing to add an empty-string alias to strain {%s}", self)
        else:
            self._aliases.add(alias)



from __future__ import annotations
from abc import ABC
from abc import abstractmethod
from .molecular_family import MolecularFamily
from .spectrum import Spectrum


class SpectrumLoaderBase(ABC):
    """Abstract base class for SpectrumLoader."""

    @property
    @abstractmethod
    def spectra(self) -> list[Spectrum]:
        """Get Spectrum objects.

        Returns:
            A sequence of Spectrum objects.
        """


class MolecularFamilyLoaderBase(ABC):
    """Abstract base class for MolecularFamilyLoader."""

    @abstractmethod
    def get_mfs(self, keep_singleton: bool) -> list[MolecularFamily]:
        """Get MolecularFamily objects.

        Args:
            keep_singleton: True to keep singleton molecular families. A
                singleton molecular family is a molecular family that contains
                only one spectrum.

        Returns:
            A sequence of MolecularFamily objects.
        """


class FileMappingLoaderBase(ABC):
    """Abstract base class for FileMappingLoader."""

    @property
    @abstractmethod
    def mappings(self) -> dict[str, list[str]]:
        """Get file mappings.

        Returns:
            A mapping from spectrum ID to the names of files where the spectrum occurs.
        """


class AnnotationLoaderBase(ABC):
    """Abstract base class for AnnotationLoader."""

    @property
    @abstractmethod
    def annotations(self) -> dict[str, dict]:
        """Get annotations.

        Returns:
            A mapping from spectrum ID to its annotations.
        """



# Copyright 2021 The NPLinker Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pickle
import numpy as np
import pylab as plt


# Method to perform the projection from the input object that
# Grimur provided
def bgc_projection(input_pickle, bgc_list, weights=[0, 1, 0]):
    # javcard, sequence, adjacency
    with open(input_pickle, "rb") as f:
        data = pickle.load(f, encoding="latin1")

    bgc_name_list = data[0]
    families = data[2]
    similarity_matrices = np.array(data[1])
    weights = np.array(weights)[:, None, None]
    d = (weights * similarity_matrices).sum(axis=0)

    L = np.diag(d) - d
    w, v = np.linalg.eig(L)

    bgc_dict = {}
    family_dict = {}
    for i, b in enumerate(bgc_name_list):
        bgc = [ba for ba in bgc_list if ba.name == b][0]
        bgc_dict[bgc] = (v[i, 0], v[i, 1])
        for family, bl in families.items():
            if b in bl:
                family_dict[bgc] = family
                break

    return bgc_dict, family_dict


def plot_and_highlight(bgc_dict, strains):
    if not isinstance(strains, list):
        strains = [strains]
    plt.figure(figsize=(10, 10))
    for b, pos in bgc_dict.items():
        plt.plot(pos[0], pos[1], "ko")
        for s in strains:
            if b.strain == s:
                plt.plot(pos[0], pos[1], "ro", markersize=10)


def plot_families(bgc_dict, family_dict, min_size=0):
    plt.figure(figsize=(10, 10))
    fam_data = {}
    for b, fam in family_dict.items():
        if fam not in fam_data:
            fam_data[fam] = []
        fam_data[fam].append(bgc_dict[b])
    for f, dat in fam_data.items():
        if len(dat) < min_size:
            plt.plot(dat[0][0], dat[0][1], "ko", markersize=5)
        else:
            x, y = zip(*dat)
            plt.plot(x, y, "o", markersize=10)


def k_means(bgc_dict, K=5, n_starts=10):
    N = len(bgc_dict)
    M = len(bgc_dict[list(bgc_dict.keys())[0]])

    bgc_list = list(bgc_dict.keys())
    X = []
    for b in bgc_list:
        X.append(bgc_dict[b])
    X = np.array(X)
    best = None
    for iteration in range(n_starts):
        mu = X[np.random.choice(N, K, replace=False), :]
        finished = False
        old_z = np.zeros((N, K))
        while not finished:
            # compute z
            some_empty = True
            while some_empty:
                di = (((X[None, :, :] - mu[:, None, :]) ** 2).sum(axis=2)).T
                z = np.zeros((N, K))
                for i, d in enumerate(di):
                    z[i, d.argmin()] = 1
                pos = np.where(z.sum(axis=0) == 0)[0]
                if len(pos) == 0:
                    some_empty = False
                for p in pos:
                    mu[p, :] = X[np.random.choice(N), :]

            # compute mu
            for k in range(K):
                mu[k, :] = (X * z[:, k][:, None]).mean(axis=0)

            if (old_z - z).sum() == 0:
                break

            old_z = z.copy()
        if iteration == 0:
            best_z = z.copy()
            best_mu = mu.copy()
        else:
            dis = 0.0
            dis_current = 0.0
            for n in range(N):
                dis += np.sqrt(((X[n, :] - mu[z[n, :].argmax(), :]) ** 2).sum())
                dis_current += np.sqrt(((X[n, :] - best_mu[best_z[n, :].argmax(), :]) ** 2).sum())
            if dis < dis_current:
                best_z = z.copy()
                best_mu = mu.copy()
                print(dis)

    families = {}
    for i, b in enumerate(bgc_list):
        families[b] = best_z[i, :].argmax()

    return families


def plot_families_strain_highlight(bgc_dict, gcf_list, strain_list):
    plt.figure(figsize=(20, 20))
    for b, pos in bgc_dict.items():
        plt.plot(pos[0], pos[1], "ko")
        plt.text(pos[0] + 0.001, pos[1] + 0.001, b.strain, fontsize=10)
    fam_success = {s: set() for s in strain_list}
    for s in strain_list:
        for f in gcf_list:
            if f.has_strain(s):
                fam_success[s].add(f)

    all_success = fam_success[strain_list[0]]
    for s in strain_list[1:]:
        all_success = all_success.intersection(fam_success[s])
    for f in all_success:
        x = []
        y = []
        for b in f.bgc_list:
            x.append(bgc_dict[b][0])
            y.append(bgc_dict[b][1])
        plt.plot(x, y, "o", markersize=10)


def hierarchical_clustering(bgc_dict, K=10):
    from genomics import GCF
    from sklearn.cluster import AgglomerativeClustering

    X = []
    bgc_list = []
    for b, pos in bgc_dict.items():
        bgc_list.append(b)
        X.append(pos)
    X = np.array(X)
    clustering = AgglomerativeClustering(n_clusters=K).fit(X)

    families = {}
    for i, b in enumerate(bgc_list):
        families[b] = clustering.labels_[i]

    gcf_dict = {i: GCF(i) for i in set(clustering.labels_)}
    for i, b in enumerate(bgc_list):
        gcf_dict[clustering.labels_[i]].add_bgc(b)
    return families, list(gcf_dict.values())



from __future__ import annotations
from collections.abc import Sequence
from functools import wraps
from typing import Union
from networkx import Graph
from tabulate import tabulate
from nplinker.genomics import GCF
from nplinker.metabolomics import MolecularFamily
from nplinker.metabolomics import Spectrum
from .score import Score
from .scoring_method import ScoringMethod


# Type aliases
Entity = Union[GCF, Spectrum, MolecularFamily]  # using Union to ensure python 3.9 compatibility
LINK_DATA = dict[str, Score]
LINK = tuple[Entity, Entity, LINK_DATA]


def validate_u(func):
    """A decorator to validate the type of the u object."""

    @wraps(func)
    def wrapper(self, u: Entity, *args, **kwargs):
        if not isinstance(u, (GCF, Spectrum, MolecularFamily)):
            raise TypeError(f"{u} is not a GCF, Spectrum, or MolecularFamily object.")

        return func(self, u, *args, **kwargs)

    return wrapper


def validate_uv(func):
    """A decorator to validate the types of the u and v objects."""

    @wraps(func)
    def wrapper(
        self,
        u: Entity,
        v: Entity,
        *args,
        **kwargs,
    ):
        if isinstance(u, GCF):
            if not isinstance(v, (Spectrum, MolecularFamily)):
                raise TypeError(f"{v} is not a Spectrum or MolecularFamily object.")
        elif isinstance(u, (Spectrum, MolecularFamily)):
            if not isinstance(v, GCF):
                raise TypeError(f"{v} is not a GCF object.")
        else:
            raise TypeError(f"{u} is not a GCF, Spectrum, or MolecularFamily object.")

        return func(self, u, v, *args, **kwargs)

    return wrapper


class LinkGraph:
    """Class to represent the links between objects in NPLinker.

    This class wraps the `networkx.Graph` class to provide a more user-friendly interface for
    working with the links.

    The links between objects are stored as edges in a graph, while the objects themselves are
    stored as nodes.

    The scoring data for each link (or link data) is stored as the key/value attributes of the edge.
    """

    def __init__(self) -> None:
        """Initialize a LinkGraph object.

        Examples:
            Create a LinkGraph object:
            >>> lg = LinkGraph()

            Display the empty LinkGraph object:
            >>> lg
            |    |   Object 1 |   Object 2 |   Metcalf Score |   Rosetta Score |
            |----|------------|------------|-----------------|-----------------|

            Add a link between a GCF and a Spectrum object:
            >>> lg.add_link(gcf, spectrum, metcalf=Score("metcalf", 1.0, {"cutoff": 0.5}))

            Display all links in LinkGraph object:
            >>> lg
            |    |     Object 1 |               Object 2 |   Metcalf Score |   Rosetta Score |
            |----|--------------|------------------------|-----------------|-----------------|
            |  1 | GCF(id=gcf1) | Spectrum(id=spectrum1) |               1 |               - |

            Get all links for a given object:
            >>> lg[gcf]
            {spectrum: {"metcalf": Score("metcalf", 1.0, {"cutoff": 0.5})}}

            Get all links in the LinkGraph:
            >>> lg.links
            [(gcf, spectrum, {"metcalf": Score("metcalf", 1.0, {"cutoff": 0.5})})]

            Check if there is a link between two objects:
            >>> lg.has_link(gcf, spectrum)
            True

            Get the link data between two objects:
            >>> lg.get_link_data(gcf, spectrum)
            {"metcalf": Score("metcalf", 1.0, {"cutoff": 0.5})}
        """
        self._g: Graph = Graph()

    def __repr__(self) -> str:
        """Return a string representation of the LinkGraph."""
        return self._get_table_repr()

    def __len__(self) -> int:
        """Get the number of objects."""
        return len(self._g)

    @validate_u
    def __getitem__(self, u: Entity) -> dict[Entity, LINK_DATA]:
        """Get all links for a given object.

        Args:
            u: the given object

        Returns:
            A dictionary of links for the given object.

        Raises:
            KeyError: if the input object is not found in the link graph.
        """
        try:
            links = self._g[u]
        except KeyError:
            raise KeyError(f"{u} not found in the link graph.")

        return {**links}  # type: ignore

    @property
    def links(
        self,
    ) -> list[LINK]:
        """Get all links.

        Returns:
            A list of tuples containing the links between objects.

        Examples:
            >>> lg.links
            [(gcf, spectrum, {"metcalf": Score("metcalf", 1.0, {"cutoff": 0.5})})]
        """
        return list(self._g.edges(data=True))

    @validate_uv
    def add_link(
        self,
        u: Entity,
        v: Entity,
        **data: Score,
    ) -> None:
        """Add a link between two objects.

        The objects `u` and `v` must be different types, i.e. one must be a GCF and the other must be
        a Spectrum or MolecularFamily.

        Args:
            u: the first object, either a GCF, Spectrum, or MolecularFamily
            v: the second object, either a GCF, Spectrum, or MolecularFamily
            data: keyword arguments. At least one scoring method and its data must be provided.
                The key must be the name of the scoring method defined in `ScoringMethod`, and the
                value is a `Score` object, e.g. `metcalf=Score("metcalf", 1.0, {"cutoff": 0.5})`.

        Examples:
            >>> lg.add_link(gcf, spectrum, metcalf=Score("metcalf", 1.0, {"cutoff": 0.5}))
        """
        # validate the data
        if not data:
            raise ValueError("At least one scoring method and its data must be provided.")
        for key, value in data.items():
            if not ScoringMethod.has_value(key):
                raise ValueError(
                    f"{key} is not a valid name of scoring method. See `ScoringMethod` for valid names."
                )
            if not isinstance(value, Score):
                raise TypeError(f"{value} is not a Score object.")

        self._g.add_edge(u, v, **data)

    @validate_uv
    def has_link(self, u: Entity, v: Entity) -> bool:
        """Check if there is a link between two objects.

        Args:
            u: the first object, either a GCF, Spectrum, or MolecularFamily
            v: the second object, either a GCF, Spectrum, or MolecularFamily

        Returns:
            True if there is a link between the two objects, False otherwise

        Examples:
            >>> lg.has_link(gcf, spectrum)
            True
        """
        return self._g.has_edge(u, v)

    @validate_uv
    def get_link_data(
        self,
        u: Entity,
        v: Entity,
    ) -> LINK_DATA | None:
        """Get the data for a link between two objects.

        Args:
            u: the first object, either a GCF, Spectrum, or MolecularFamily
            v: the second object, either a GCF, Spectrum, or MolecularFamily

        Returns:
            A dictionary of scoring methods and their data for the link between the two objects, or
            None if there is no link between the two objects.

        Examples:
            >>> lg.get_link_data(gcf, spectrum)
            {"metcalf": Score("metcalf", 1.0, {"cutoff": 0.5})}
        """
        return self._g.get_edge_data(u, v)  # type: ignore

    def filter(self, u_nodes: Sequence[Entity], v_nodes: Sequence[Entity] = [], /) -> LinkGraph:
        """Return a new LinkGraph object with the filtered links between the given objects.

        The new LinkGraph object will only contain the links between `u_nodes` and `v_nodes`.

        If `u_nodes` or `v_nodes` is empty, the new LinkGraph object will contain the links for
        the given objects in `v_nodes` or `u_nodes`, respectively. If both are empty, return an
        empty LinkGraph object.

        Note that not all objects in `u_nodes` and `v_nodes` need to be present in the original
        LinkGraph.

        Args:
            u_nodes: a sequence of objects used as the first object in the links
            v_nodes: a sequence of objects used as the second object in the links

        Returns:
            A new LinkGraph object with the filtered links between the given objects.

        Examples:
            Filter the links for `gcf1` and `gcf2`:
            >>> new_lg = lg.filter([gcf1, gcf2])
            Filter the links for `spectrum1` and `spectrum2`:
            >>> new_lg = lg.filter([spectrum1, spectrum2])
            Filter the links between two lists of objects:
            >>> new_lg = lg.filter([gcf1, gcf2], [spectrum1, spectrum2])
        """
        lg = LinkGraph()

        # exchange u_nodes and v_nodes if u_nodes is empty but v_nodes not
        if len(u_nodes) == 0 and len(v_nodes) != 0:
            u_nodes = v_nodes
            v_nodes = []

        if len(v_nodes) == 0:
            for u in u_nodes:
                self._filter_one_node(u, lg)

        for u in u_nodes:
            for v in v_nodes:
                self._filter_two_nodes(u, v, lg)

        return lg

    @validate_u
    def _filter_one_node(self, u: Entity, lg: LinkGraph) -> None:
        """Filter the links for a given object and add them to the new LinkGraph object."""
        try:
            links = self[u]
        except KeyError:
            pass
        else:
            for node2, value in links.items():
                lg.add_link(u, node2, **value)

    @validate_uv
    def _filter_two_nodes(self, u: Entity, v: Entity, lg: LinkGraph) -> None:
        """Filter the links between two objects and add them to the new LinkGraph object."""
        link_data = self.get_link_data(u, v)
        if link_data is not None:
            lg.add_link(u, v, **link_data)

    def _get_table_repr(self) -> str:
        """Generate a table representation of the LinkGraph.

        The table is truncated to 60 links.
        """
        headers = ["", "Object 1", "Object 2", "Metcalf Score", "Rosetta Score"]
        table_data = []
        display_limit = 60

        for index, (u, v, data) in enumerate(self.links, start=1):
            metcalf_score = data.get("metcalf")
            rosetta_score = data.get("rosetta")

            row = [
                index,
                str(u if isinstance(u, GCF) else v),
                str(v if isinstance(u, GCF) else u),
                f"{metcalf_score.value:.2f}" if metcalf_score else "-",
                f"{rosetta_score.value:.2f}" if rosetta_score else "-",
            ]
            table_data.append(row)

            if index == display_limit:
                break

        table = tabulate(table_data, headers=headers, tablefmt="github", stralign="right")

        if len(self.links) > display_limit:
            truncated_info = f"...\n[ {len(self.links)} links ]"
            return f"{table}\n{truncated_info}"

        return table



from __future__ import annotations
from typing import TYPE_CHECKING
from nplinker.strain import Strain
from nplinker.strain import StrainCollection


if TYPE_CHECKING:
    from .spectrum import Spectrum


class MolecularFamily:
    """Class to model molecular family.

    Attributes:
        id: Unique id for the molecular family.
        spectra_ids: Set of spectrum ids in the molecular family.
        spectra: Set of Spectrum objects in the molecular family.
        strains: StrainCollection object that contains strains in the molecular family.
    """

    def __init__(self, id: str):
        """Initialize the MolecularFamily.

        Args:
            id: Unique id for the molecular family.
        """
        self.id: str = id
        self.spectra_ids: set[str] = set()
        self._spectra: set[Spectrum] = set()
        self._strains: StrainCollection = StrainCollection()

    def __str__(self) -> str:
        return (
            f"MolecularFamily(id={self.id}, #Spectrum_objects={len(self._spectra)}, "
            f"#spectrum_ids={len(self.spectra_ids)}, #strains={len(self._strains)})"
        )

    def __repr__(self) -> str:
        return str(self)

    def __eq__(self, other) -> bool:
        if isinstance(other, MolecularFamily):
            return self.id == other.id
        return NotImplemented

    def __hash__(self) -> int:
        return hash(self.id)

    def __reduce__(self) -> tuple:
        """Reduce function for pickling."""
        return (self.__class__, (self.id,), self.__dict__)

    @property
    def spectra(self) -> set[Spectrum]:
        """Get Spectrum objects in the molecular family."""
        return self._spectra

    @property
    def strains(self) -> StrainCollection:
        """Get strains in the molecular family."""
        return self._strains

    def add_spectrum(self, spectrum: Spectrum) -> None:
        """Add a Spectrum object to the molecular family.

        Args:
            spectrum: `Spectrum` object to add to the molecular family.
        """
        self._spectra.add(spectrum)
        self.spectra_ids.add(spectrum.id)
        self._strains = self._strains + spectrum.strains
        # add the molecular family to the spectrum
        spectrum.family = self

    def detach_spectrum(self, spectrum: Spectrum) -> None:
        """Remove a Spectrum object from the molecular family.

        Args:
            spectrum: `Spectrum` object to remove from the molecular family.
        """
        self._spectra.remove(spectrum)
        self.spectra_ids.remove(spectrum.id)
        self._strains = self._update_strains()
        # remove the molecular family from the spectrum
        spectrum.family = None

    def has_strain(self, strain: Strain) -> bool:
        """Check if the given strain exists.

        Args:
            strain: `Strain` object.

        Returns:
            True when the given strain exists.
        """
        return strain in self._strains

    def is_singleton(self) -> bool:
        """Check if the molecular family contains only one spectrum.

        Returns:
            True when the molecular family has only one spectrum.
        """
        return len(self.spectra_ids) == 1

    def _update_strains(self) -> StrainCollection:
        """Update strains in the molecular family.

        The strains are re-extracted from the existing spectra in the molecular family. This method
        is mainly used when the spectra are updated (e.g. remove a spectrum from the molecular
        family).

        Returns:
            Updated StrainCollection object.
        """
        strains = StrainCollection()
        for spectrum in self._spectra:
            strains = strains + spectrum.strains
        return strains



from __future__ import annotations
import logging
from enum import Enum
from typing import TYPE_CHECKING
from typing import Any
from typing import Union
from typing import overload
import numpy as np
import pandas as pd
from scipy.stats import hypergeom
from nplinker.genomics import GCF
from nplinker.metabolomics import MolecularFamily
from nplinker.metabolomics import Spectrum
from .abc import ScoringBase
from .link_graph import LinkGraph
from .link_graph import Score
from .scoring_method import ScoringMethod
from .utils import get_presence_gcf_strain
from .utils import get_presence_mf_strain
from .utils import get_presence_spec_strain


if TYPE_CHECKING:
    from nplinker.nplinker import NPLinker


logger = logging.getLogger(__name__)


class LinkType(Enum):
    """Enum class for link types."""

    SPEC_GCF = "spec-gcf"
    MF_GCF = "mf-gcf"


Entity = Union[GCF, Spectrum, MolecularFamily]


class MetcalfScoring(ScoringBase):
    """Metcalf scoring method.

    Attributes:
        name: The name of this scoring method, set to a fixed value `metcalf`.
        npl: The NPLinker object.
        CACHE: The name of the cache file to use for storing the MetcalfScoring.

        presence_gcf_strain: A DataFrame to store presence of gcfs with respect to strains.
            The index of the DataFrame are the GCF objects and the columns are Strain objects.
            The values are 1 where the gcf occurs in the strain, 0 otherwise.
        presence_spec_strain: A DataFrame to store presence of spectra with respect to strains.
            The index of the DataFrame are the Spectrum objects and the columns are Strain objects.
            The values are 1 where the spectrum occurs in the strain, 0 otherwise.
        presence_mf_strain: A DataFrame to store presence of molecular families with respect to strains.
            The index of the DataFrame are the MolecularFamily objects and the columns are Strain objects.
            The values are 1 where the molecular family occurs in the strain, 0 otherwise.

        raw_score_spec_gcf: A DataFrame to store the raw Metcalf scores for spectrum-gcf links.
            The columns are "spec", "gcf" and "score":

            - The "spec" and "gcf" columns contain the Spectrum and GCF objects respectively,
            - The "score" column contains the raw Metcalf scores.

        raw_score_mf_gcf: A DataFrame to store the raw Metcalf scores for molecular family-gcf links.
            The columns are "mf", "gcf" and "score":

            - The "mf" and "gcf" columns contain the MolecularFamily and GCF objects respectively,
            - the "score" column contains the raw Metcalf scores.

        metcalf_mean: A numpy array to store the mean value used for standardising Metcalf scores.
            The array has shape (n_strains+1, n_strains+1), where n_strains is the number of strains.
        metcalf_std: A numpy array to store the standard deviation value used for standardising
            Metcalf scores. The array has shape (n_strains+1, n_strains+1), where n_strains is the
            number of strains.
    """

    name = ScoringMethod.METCALF.value
    npl: NPLinker | None = None
    CACHE: str = "cache_metcalf_scoring.pckl"
    metcalf_weights: tuple[int, int, int, int] = (10, -10, 0, 1)

    # index: gcf/spec/mf ids, columns: strain ids, value: 0/1
    presence_gcf_strain: pd.DataFrame = pd.DataFrame()
    presence_spec_strain: pd.DataFrame = pd.DataFrame()
    presence_mf_strain: pd.DataFrame = pd.DataFrame()

    raw_score_spec_gcf: pd.DataFrame = pd.DataFrame(columns=["spec", "gcf", "score"])
    raw_score_mf_gcf: pd.DataFrame = pd.DataFrame(columns=["mf", "gcf", "score"])

    metcalf_mean: np.ndarray | None = None
    metcalf_std: np.ndarray | None = None

    @classmethod
    def setup(cls, npl: NPLinker) -> None:
        """Setup the MetcalfScoring object.

        This method is only called once to setup the MetcalfScoring object.

        Args:
            npl: The NPLinker object.
        """
        if cls.npl is not None:
            logger.info("MetcalfScoring.setup already called, skipping.")
            return

        logger.info(
            f"MetcalfScoring.setup starts: #bgcs={len(npl.bgcs)}, #gcfs={len(npl.gcfs)}, "
            f"#spectra={len(npl.spectra)}, #mfs={len(npl.mfs)}, #strains={npl.strains}"
        )
        cls.npl = npl

        # calculate presence of gcfs/spectra/mfs with respect to strains
        cls.presence_gcf_strain = get_presence_gcf_strain(npl.gcfs, npl.strains)
        cls.presence_spec_strain = get_presence_spec_strain(npl.spectra, npl.strains)
        cls.presence_mf_strain = get_presence_mf_strain(npl.mfs, npl.strains)

        # calculate raw Metcalf scores for spec-gcf links
        raw_score_spec_gcf = cls._calc_raw_score(
            cls.presence_spec_strain, cls.presence_gcf_strain, cls.metcalf_weights
        )
        cls.raw_score_spec_gcf = raw_score_spec_gcf.reset_index().melt(id_vars="index")
        cls.raw_score_spec_gcf.columns = ["spec", "gcf", "score"]  # type: ignore

        # calculate raw Metcalf scores for spec-gcf links
        raw_score_mf_gcf = cls._calc_raw_score(
            cls.presence_mf_strain, cls.presence_gcf_strain, cls.metcalf_weights
        )
        cls.raw_score_mf_gcf = raw_score_mf_gcf.reset_index().melt(id_vars="index")
        cls.raw_score_mf_gcf.columns = ["mf", "gcf", "score"]  # type: ignore

        # calculate mean and std for standardising Metcalf scores
        cls.metcalf_mean, cls.metcalf_std = cls._calc_mean_std(
            len(npl.strains), cls.metcalf_weights
        )

        logger.info("MetcalfScoring.setup completed")

    @overload
    def get_links(self, *objects: GCF, **parameters: Any) -> LinkGraph: ...
    @overload
    def get_links(self, *objects: Spectrum, **parameters: Any) -> LinkGraph: ...
    @overload
    def get_links(self, *objects: MolecularFamily, **parameters: Any) -> LinkGraph: ...

    def get_links(self, *objects, **parameters):
        """Get links for the given objects.

        Args:
            objects: The objects to get links for. All objects must be of the same type, i.e. `GCF`,
                `Spectrum` or `MolecularFamily` type.
                If no objects are provided, all detected objects (`npl.gcfs`) will be used.
            parameters: The scoring parameters to use for the links.
                The parameters are:

                - `cutoff`: The minimum score to consider a link (≥cutoff). Default is 0.
                - `standardised`: Whether to use standardised scores. Default is False.

        Returns:
            The [`LinkGraph`][nplinker.scoring.LinkGraph] object containing the links involving the
                input objects with the Metcalf scores.

        Raises:
            TypeError: If the input objects are not of the same type or the object type is invalid.
        """
        # validate input objects
        if len(objects) == 0:
            objects = self.npl.gcfs
        # check if all objects are of the same type
        types = {type(i) for i in objects}
        if len(types) > 1:
            raise TypeError("Input objects must be of the same type.")
        # check if the object type is valid
        obj_type = next(iter(types))
        if obj_type not in (GCF, Spectrum, MolecularFamily):
            raise TypeError(
                f"Invalid type {obj_type}. Input objects must be GCF, Spectrum or MolecularFamily objects."
            )

        # validate scoring parameters
        self._cutoff: float = parameters.get("cutoff", 0)
        self._standardised: bool = parameters.get("standardised", False)
        parameters.update({"cutoff": self._cutoff, "standardised": self._standardised})

        logger.info(
            f"MetcalfScoring: #objects={len(objects)}, type={obj_type}, cutoff={self._cutoff}, "
            f"standardised={self._standardised}"
        )
        if not self._standardised:
            scores_list = self._get_links(*objects, obj_type=obj_type, score_cutoff=self._cutoff)
        else:
            if self.metcalf_mean is None or self.metcalf_std is None:
                raise ValueError(
                    "MetcalfScoring.metcalf_mean and metcalf_std are not set. Run MetcalfScoring.setup first."
                )
            # use negative infinity as the score cutoff to ensure we get all links
            scores_list = self._get_links(*objects, obj_type=obj_type, score_cutoff=-np.inf)
            scores_list = self._calc_standardised_score(scores_list)

        links = LinkGraph()
        for score_df in scores_list:
            for row in score_df.itertuples(index=False):  # row has attributes: spec/mf, gcf, score
                met = row.spec if score_df.name == LinkType.SPEC_GCF else row.mf
                links.add_link(
                    row.gcf,
                    met,
                    metcalf=Score(self.name, row.score, parameters),
                )

        logger.info(f"MetcalfScoring: completed! Found {len(links.links)} links in total.")
        return links

    # TODO CG: refactor this method
    def format_data(self, data):
        """Format the data for display."""
        # for metcalf the data will just be a floating point value (i.e. the score)
        return f"{data:.4f}"

    # TODO CG: refactor this method
    def sort(self, objects, reverse=True):
        """Sort the objects based on the score."""
        # sort based on score
        return sorted(objects, key=lambda objlink: objlink[self], reverse=reverse)

    @staticmethod
    def _calc_raw_score(
        p1: pd.DataFrame, p2: pd.DataFrame, weights: tuple[int, int, int, int]
    ) -> pd.DataFrame:
        """Calculate non-standardised Metcalf scores.

        Args:
            p1: A DataFrame containing the presence of objects in strains.
            p2: A DataFrame containing the presence of objects in strains.
            weights: The weights to use for Metcalf scoring.

        Returns:
            A DataFrame containing the non-standardised Metcalf scores.
        """
        nop1 = 1 - p1
        nop2 = 1 - p2

        # calculate co-presence
        p1_p2 = p1.dot(p2.T)
        p1_nop2 = p1.dot(nop2.T)
        nop1_p2 = nop1.dot(p2.T)
        nop1_nop2 = nop1.dot(nop2.T)

        # calculate weighted sum
        score = (
            p1_p2 * weights[0]
            + p1_nop2 * weights[1]
            + nop1_p2 * weights[2]
            + nop1_nop2 * weights[3]
        )

        return score

    @staticmethod
    def _calc_mean_std(
        n_strains: int, weights: tuple[int, int, int, int]
    ) -> tuple[np.ndarray, np.ndarray]:
        """Calculate the mean and standard deviation for Metcalf scoring.

        Args:
            n_strains: The number of strains.
            weights: The weights to use for Metcalf scoring.

        Returns:
            Two numpy arrays containing the mean and standard deviation values for Metcalf scoring.
            The arrays have shape (n_strains+1, n_strains+1).
        """
        sz = (n_strains + 1, n_strains + 1)
        mean = np.zeros(sz)
        variance = np.zeros(sz)
        for n in range(n_strains + 1):
            for m in range(n_strains + 1):
                max_overlap = min(n, m)
                min_overlap = max(0, n + m - n_strains)
                expected_value = 0
                expected_sq = 0
                for o in range(min_overlap, max_overlap + 1):
                    o_prob = hypergeom.pmf(o, n_strains, n, m)
                    # compute metcalf for n strains in type 1 and m in gcf
                    score = o * weights[0]
                    score += weights[1] * (n - o)
                    score += weights[2] * (m - o)
                    score += weights[3] * (n_strains - (n + m - o))
                    expected_value += o_prob * score
                    expected_sq += o_prob * (score**2)
                mean[n, m] = expected_value
                expected_sq = expected_sq - expected_value**2
                if expected_sq < 1e-09:
                    expected_sq = 1
                variance[n, m] = expected_sq
        return mean, np.sqrt(variance)

    def _get_links(
        self,
        *objects: Entity,
        obj_type: Entity,
        score_cutoff: float = 0,
    ) -> list[pd.DataFrame]:
        """Get links and scores for the given objects.

        Args:
            objects: A list of GCF, Spectrum or MolecularFamily objects and all objects must be of
                the same type.
            obj_type: The type of the objects.
            score_cutoff: Minimum score to consider a link (≥score_cutoff). Default is 0.

        Returns:
            List of data frames containing the ids of the linked objects and the score.

            The data frame is named by link types, see `LinkType`. It has column names of
            ['spec', 'gcf', 'score'] or ['mf', 'gcf', 'score'] depending on the link type:

            - the 'spec', 'mf' or 'gcf' column contains the Spectrum, MolecularFamily or GCF objects,
            - the 'score' column contains the scores.
        """
        links = []
        col_name = (
            "spec" if obj_type == Spectrum else "mf" if obj_type == MolecularFamily else "gcf"
        )

        # spec-gcf link
        if obj_type in (GCF, Spectrum):
            df = self.raw_score_spec_gcf[
                self.raw_score_spec_gcf[col_name].isin(objects)
                & (self.raw_score_spec_gcf["score"] >= score_cutoff)
            ]
            df.name = LinkType.SPEC_GCF
            links.append(df)

        # mf-gcf link
        if obj_type in (GCF, MolecularFamily):
            df = self.raw_score_mf_gcf[
                self.raw_score_mf_gcf[col_name].isin(objects)
                & (self.raw_score_mf_gcf["score"] >= score_cutoff)
            ]
            df.name = LinkType.MF_GCF
            links.append(df)

        return links

    def _calc_standardised_score(self, raw_scores: list[pd.DataFrame]) -> list[pd.DataFrame]:
        """Calculate standardised Metcalf scores.

        Args:
            raw_scores: A list of DataFrames containing the raw Metcalf scores.

        Returns:
            A list of DataFrames containing the standardised Metcalf scores.
        """
        standardised_scores = []
        for raw_score_df in raw_scores:
            # create a new DataFrame to store the standardised scores, with the same columns
            # and name as the raw score DataFrame
            standardised_score_df = pd.DataFrame(columns=raw_score_df.columns)

            for row in raw_score_df.itertuples(index=False):
                met = row.spec if raw_score_df.name == LinkType.SPEC_GCF else row.mf
                n_gcf_strains = len(row.gcf.strains)  # type: ignore
                n_met_strains = len(met.strains)  # type: ignore

                mean = self.metcalf_mean[n_met_strains][n_gcf_strains]  # type: ignore
                sqrt = self.metcalf_std[n_met_strains][n_gcf_strains]  # type: ignore

                z_score = (row.score - mean) / sqrt

                if z_score >= self._cutoff:
                    # add the row to the standardised score DataFrame with the z-score as score value
                    data = {col_name: getattr(row, col_name) for col_name in raw_score_df.columns}
                    data["score"] = z_score
                    new_row = pd.DataFrame(data, index=[0])
                    standardised_score_df = pd.concat(
                        (standardised_score_df, new_row), ignore_index=True
                    )
                else:
                    continue

            standardised_score_df.name = raw_score_df.name
            standardised_scores.append(standardised_score_df)

        return standardised_scores



# Copyright 2021 The NPLinker Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy
from numba import jit


class MSSpectrum:
    def __init__(self, mgf_dict=None, spec=None):
        self.compound = None
        self.formula = None
        self.ionisation = None
        self.raw_parentmass = None
        self.filename = None
        self.id = None

        self.raw_spectrum = None
        self.output_spectrum = None

        # self.filter is a function that can be used
        # to filter the raw spectrum, e.g. denoising,
        # removal of adduct, etc.
        self.filter = None
        self.normalise = False

        self.correct_for_ionisation = False

        if mgf_dict is not None:
            self.init_from_mgf(mgf_dict)

        if spec is not None:
            self.init_from_spec(spec)

    def init_from_spec(self, spec):
        self.id = spec.id
        self.raw_parentmass = spec.precursor_mz
        self.raw_spectrum = spec.peaks
        # TODO this is a temporary default for the Crusemann data
        # should check for it in the mgf in metabolomics.py and store
        # in the Spectrum object if found
        # (TODO what is the MGF field name for this??)
        self.ionisation = "[M+H]+"

    def init_from_mgf(self, mgf_dict):
        self.output_spectrum = None

        self.filename = mgf_dict["params"]["filename"]
        self.compound = None
        self.formula = None
        self.ionisation = mgf_dict["params"]["charge"]
        self.raw_parentmass = mgf_dict["params"]["pepmass"][0]
        self.inchi = None
        self.id = None
        if "smiles" in mgf_dict["params"]:
            self.smiles = mgf_dict["params"]["smiles"]

        spec = []
        for a in zip(mgf_dict["m/z array"], mgf_dict["intensity array"]):
            spec.append(a)

        self.raw_spectrum = numpy.array(spec)

    def load(self, filename):
        self.output_spectrum = None
        self.filename = filename
        spectrum = []
        with open(filename) as f:
            for line in f.readlines():
                line = line.strip()
                if len(line) == 0:
                    pass
                elif line.startswith(">compound"):
                    self.compound = strip_leading(line)
                elif line.startswith(">formula"):
                    self.formula = strip_leading(line)
                elif line.startswith(">ionization"):
                    self.ionisation = strip_leading(line)
                elif line.startswith(">parentmass"):
                    self.raw_parentmass = float(strip_leading(line))
                elif line.startswith(">"):
                    pass
                elif line.startswith("#inchi"):
                    self.inchi = strip_leading(line)
                elif line.startswith("#SpectrumID"):
                    self.id = strip_leading(line)
                elif line.startswith("#"):
                    pass
                else:
                    mass, charge = line.split()
                    mass = float(mass)
                    charge = float(charge)
                    spectrum.append((mass, charge))
        self.raw_spectrum = numpy.array(spectrum)

    @property
    def parentmass(self):
        if self.correct_for_ionisation:
            return self.raw_parentmass - self.ionisation_mass
        else:
            return self.raw_parentmass

    @property
    def spectrum(self):
        if self.normalise:
            return _normalise_spectrum(self.unnormalised_spectrum)
        else:
            return self.unnormalised_spectrum

    @property
    def unnormalised_spectrum(self):
        if self.filter is None:
            if self.correct_for_ionisation:
                return self.shifted_spectrum
            else:
                return self.raw_spectrum
        else:
            if self.output_spectrum is None:
                filtered_spectrum = self.filter(self)
                if len(filtered_spectrum) != 0:
                    self.output_spectrum = self.filter(self)
                else:
                    self.output_spectrum = self.shifted_spectrum
            return self.output_spectrum

    @property
    def shifted_spectrum(self):
        return self.raw_spectrum - [self.ionisation_mass, 0]

    @property
    def ionisation_mass(self):
        return IONISATION_MASSES[self.ionisation]


PROTON_MASS = 1.00727645199076
IONISATION_MASSES = {
    "[M+H]+": PROTON_MASS,
    "[M+H-H2O]+": PROTON_MASS - 18.01056468638,
    "[M+K]+": 38.963158,
    "[M+Na]+": 22.989218,
}


def _normalise_spectrum(spectrum, peak=100.0):
    _, max_peak = numpy.max(spectrum, axis=0)
    return spectrum * [1, peak / max_peak]


def strip_leading(line):
    return " ".join(line.split()[1:])


def _ppk(i_peaks, j_peaks, sm, si):
    X1 = i_peaks
    X2 = j_peaks
    # N1 = numpy.size(X1, 0); N2 = numpy.size(X2, 0)
    N1 = X1.shape[0]
    N2 = X2.shape[0]
    if N1 == 0 or N2 == 0:
        raise Exception("[ERROR]:No peaks when computing the kernel.(try not clean the peaks)")
    constant = 1.0 / (N1 * N2) * 0.25 / (numpy.pi * numpy.sqrt(sm * si))
    mass_term = (
        1.0
        / sm
        * numpy.power(
            numpy.kron(X1[:, 0].flatten(), numpy.ones(N2))
            - numpy.kron(numpy.ones(N1), X2[:, 0].flatten()),
            2,
        )
    )
    inte_term = (
        1.0
        / si
        * numpy.power(
            numpy.kron(X1[:, 1].flatten(), numpy.ones(N2))
            - numpy.kron(numpy.ones(N1), X2[:, 1].flatten()),
            2,
        )
    )
    return constant * numpy.sum(numpy.exp(-0.25 * (mass_term + inte_term)))


def ppk(*args):
    # t0 = time.time()
    # a = ppk_loop(*args)
    # = _ppk(*args)
    # t1 = time.time()
    b = ppk_limit(*args)
    # t2 = time.time()
    # print(t1-t0, t2-t1)
    return b


@jit(nopython=True)
def ppk_loop(spectrum_1, spectrum_2, sigma_mass, sigma_int):
    # the inputs are really sigma^2, though
    # sigma_mass = 0.00001
    # sigma_int = 100000
    sigma_array = numpy.array([[sigma_mass, 0], [0, sigma_int]])
    sigma_inv = numpy.linalg.inv(sigma_array)
    len_1 = spectrum_1.shape[0]
    len_2 = spectrum_2.shape[0]
    constant_term = 1.0 / (len_1 * len_2 * 4 * numpy.pi * numpy.sqrt(sigma_mass * sigma_int))
    sum_term = 0
    # for p_1, p_2 in itertools.product(spectrum_1, spectrum_2):
    for p_1_idx in range(len_1):
        p_1 = spectrum_1[p_1_idx, :]
        for p_2_idx in range(len_2):
            p_2 = spectrum_2[p_2_idx, :]
            d = p_1 - p_2
            sum_term += numpy.exp(-0.25 * numpy.sum(d * sigma_inv * d))
    # print(sum_term)
    # print(numpy.sum(sum_term))
    return constant_term * sum_term


@jit(nopython=True)
def ppk_limit(spectrum_1, spectrum_2, sigma_mass, sigma_int):
    # the inputs are really sigma^2, though
    # sigma_mass = 0.00001
    # sigma_int = 100000
    sigma_array = numpy.array([[sigma_mass, 0], [0, sigma_int]])
    sigma_inv = numpy.linalg.inv(sigma_array)
    len_1 = spectrum_1.shape[0]
    len_2 = spectrum_2.shape[0]
    constant_term = 1.0 / (len_1 * len_2 * 4 * numpy.pi * numpy.sqrt(sigma_mass * sigma_int))
    sum_term = 0

    tol = 5 * numpy.sqrt(sigma_mass)

    for p_1_idx, p_2_idx in find_pairs(spectrum_1, spectrum_2, tol):
        p_1 = spectrum_1[p_1_idx, :]
        p_2 = spectrum_2[p_2_idx, :]
        d = p_1 - p_2
        sum_term += numpy.exp(-0.25 * numpy.sum(d * sigma_inv * d))
    # print(sum_term)
    # print(numpy.sum(sum_term))
    # print(constant_term, sum_term)
    return constant_term * sum_term


@jit(nopython=True)
def find_pairs(spec1, spec2, tol, shift=0):
    matching_pairs = []
    spec2_lowpos = 0
    spec2_length = len(spec2)

    for idx in range(len(spec1)):
        mz, intensity = spec1[idx, :]
        while spec2_lowpos < spec2_length and spec2[spec2_lowpos][0] + shift < mz - tol:
            spec2_lowpos += 1
        if spec2_lowpos == spec2_length:
            break
        spec2_pos = spec2_lowpos
        while spec2_pos < spec2_length and spec2[spec2_pos][0] + shift < mz + tol:
            matching_pairs.append((idx, spec2_pos))
            spec2_pos += 1

    return matching_pairs


def ppk_nloss(spec1, spec2, prec1, prec2, sigma_mass, sigma_int):
    spec1_loss = ([prec1, 0] - spec1) * [1, -1]
    spec2_loss = ([prec2, 0] - spec2) * [1, -1]
    k_nloss = ppk(spec1_loss[::-1], spec2_loss[::-1], sigma_mass, sigma_int)
    return k_nloss



# Copyright 2021 The NPLinker Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
from sortedcontainers import SortedList
from nplinker.metabolomics.gnps import GNPSSpectrumLoader
from .rosetta_functions import fast_cosine


logger = logging.getLogger(__name__)


class SpecLib:
    def __init__(self, mgf_file):
        self.mgf_file = mgf_file
        self.spectra = []

    def _load_mgf(self):
        self.spectra = GNPSSpectrumLoader(self.mgf_file).spectra()

        self.sort()

    def sort(self):
        # make a sorted list for quick precursor matching
        self.sorted_spectra = [s for s in self.spectra]
        self.sorted_spectra.sort()

    def get_n_spec(self):
        return len(self.spectra)

    def get_ids(self):
        return list(s.id for s in self.spectra)

    def get_n_peaks(self):
        return [len(s.peaks) for s in self.spectra]

    def filter(self):
        # top_k_filter
        n_done = 0
        for spec in self.spectra:
            self._keep_top_k(spec)
            n_done += 1
            if n_done % 100 == 0:
                logger.info(
                    "SpecLib filtered {}/{}, {:.2f}%".format(
                        n_done, len(self.spectra), 100 * (n_done / len(self.spectra))
                    )
                )

    def spectral_match(
        self,
        query,
        scoring_function=fast_cosine,
        ms2_tol=0.2,
        min_match_peaks=1,
        ms1_tol=0.2,
        score_thresh=0.7,
    ):
        candidates = self._candidates(self.sorted_spectra, query.precursor_mz, ms1_tol)
        hits = []
        for c in candidates:
            sc, _ = scoring_function(query, c, ms2_tol, min_match_peaks)
            if sc >= score_thresh:
                hits.append((c.gnps_id, sc))
        return hits

    def _candidates(self, mz_list, query_mz, ms1_tol):
        pmz_list = SortedList([m.precursor_mz for m in mz_list])
        lower = query_mz - ms1_tol
        upper = query_mz + ms1_tol
        start = pmz_list.bisect(lower)
        end = pmz_list.bisect(upper)
        return mz_list[start:end]

    # from molnet repo
    def _keep_top_k(self, spec, k=6, mz_range=50):
        # only keep peaks that are in the top k in += mz_range
        start_pos = 0
        new_mz = []
        new_intensities = []
        for mz, intensity in spec.peaks:
            while spec.peaks[start_pos][0] < mz - mz_range:
                start_pos += 1
            end_pos = start_pos

            n_bigger = 0
            while end_pos < len(spec.peaks) and spec.peaks[end_pos][0] <= mz + mz_range:
                if spec.peaks[end_pos][1] > intensity:
                    n_bigger += 1
                end_pos += 1

            if n_bigger < k:
                new_mz.append(mz)
                new_intensities.append(intensity)

        spec.mz = new_mz
        spec.intensity = new_intensities



from __future__ import annotations
import logging
from typing import TYPE_CHECKING
from nplinker.strain import Strain
from nplinker.strain import StrainCollection


if TYPE_CHECKING:
    from .bgc import BGC

logger = logging.getLogger(__name__)


class GCF:
    """Class to model gene cluster family (GCF).

    GCF is a group of similar BGCs and generated by clustering BGCs with
    tools such as BiG-SCAPE and BiG-SLICE.

    Attributes:
        id: id of the GCF object.
        bgc_ids: a set of BGC ids that belongs to the GCF.
        bigscape_class: BiG-SCAPE's BGC class.
            BiG-SCAPE's BGC classes are similar to those defined in MiBIG
            but have more categories (7 classes), including:

            - NRPS
            - PKS-NRP_Hybrids
            - PKSI
            - PKSother
            - RiPPs
            - Saccharides
            - Terpene

            For BGC falls outside of these categories, the value is "Others".

            Default is None, which means the class is unknown.

            More details see:
            https://doi.org/10.1038%2Fs41589-019-0400-9.
    """

    def __init__(self, id: str, /) -> None:
        """Initialize the GCF object.

        Args:
            id: id of the GCF object.

        Examples:
            >>> gcf = GCF("Unique_GCF_ID")
            >>> gcf.id
            'Unique_GCF_ID'
        """
        self.id = id
        self.bgc_ids: set[str] = set()
        self.bigscape_class: str | None = None
        self._bgcs: set[BGC] = set()
        self._strains: StrainCollection = StrainCollection()

    def __str__(self) -> str:
        return (
            f"GCF(id={self.id}, #BGC_objects={len(self.bgcs)}, #bgc_ids={len(self.bgc_ids)},"
            f"#strains={len(self._strains)})."
        )

    def __repr__(self) -> str:
        return str(self)

    def __eq__(self, other) -> bool:
        if isinstance(other, GCF):
            return self.id == other.id and self.bgcs == other.bgcs
        return NotImplemented

    def __hash__(self) -> int:
        """Hash function for GCF.

        Note that GCF class is a mutable container. We only hash the GCF id to
        avoid the hash value changes when `self._bgcs` is updated.
        """
        return hash(self.id)

    def __reduce__(self) -> tuple:
        """Reduce function for pickling."""
        return (self.__class__, (self.id,), self.__dict__)

    @property
    def bgcs(self) -> set[BGC]:
        """Get the BGC objects."""
        return self._bgcs

    @property
    def strains(self) -> StrainCollection:
        """Get the strains in the GCF."""
        return self._strains

    def add_bgc(self, bgc: BGC) -> None:
        """Add a BGC object to the GCF."""
        bgc.parents.add(self)
        self._bgcs.add(bgc)
        self.bgc_ids.add(bgc.id)
        if bgc.strain is not None:
            self._strains.add(bgc.strain)
        else:
            logger.warning("No strain specified for the BGC %s", bgc.id)

    def detach_bgc(self, bgc: BGC) -> None:
        """Remove a child BGC object."""
        bgc.parents.remove(self)
        self._bgcs.remove(bgc)
        self.bgc_ids.remove(bgc.id)
        if bgc.strain is not None:
            for other_bgc in self._bgcs:
                if other_bgc.strain == bgc.strain:
                    return
            self._strains.remove(bgc.strain)

    def has_strain(self, strain: Strain) -> bool:
        """Check if the given strain exists.

        Args:
            strain: `Strain` object.

        Returns:
            True when the given strain exist.
        """
        return strain in self._strains

    def has_mibig_only(self) -> bool:
        """Check if the GCF's children are only MIBiG BGCs.

        Returns:
            True if `GCF.bgc_ids` are only MIBiG BGC ids.
        """
        return all(map(lambda id: id.startswith("BGC"), self.bgc_ids))

    def is_singleton(self) -> bool:
        """Check if the GCF contains only one BGC.

        Returns:
            True if `GCF.bgc_ids` contains only one BGC id.
        """
        return len(self.bgc_ids) == 1



# Copyright 2021 The NPLinker Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import os
from collections import Counter
from collections import defaultdict
import pandas as pd


logger = logging.getLogger(__name__)


class ClassMatches:
    """Holds all info concerning class matches (based on known bgc-structure links in MIBiG)."""

    def __init__(self, mibig_classes_file):
        """Read mibig classes file and convert into count and scoring tables.

        Args:
            mibig_classes_file: str, filepath of the MIBiG data as made by
            https://github.com/louwenjjr/mibig_classifications
        """
        self._mibig_classes_file = mibig_classes_file
        (
            self._class_matches,
            self._class_matches_counts,
            self._bgc_class_names,
            self._chem_class_names,
        ) = ({}, {}, [], [])

        if os.path.isfile(self._mibig_classes_file):
            self._read_mibig_classes()
            self._get_class_counts()
            self._get_scoring_tables()
            logger.info("Loaded MIBiG classes, and class matching tables")
        else:
            logger.warn("MIBiG classes not found, class matches not loaded!")

        pd.options.display.float_format = "{:,.3f}".format  # adjust pd formatting

        self._bigscape_mibig_conversion = {
            "PKSI": "Polyketide",
            "PKSother": "Polyketide",
            "NRPS": "NRP",
            "RiPPs": "RiPP",
            "Saccharides": "Saccharide",
            "Others": "Other",
            "Terpene": "Terpene",
            "PKS-NRP_Hybrids": "PKS-NRP_Hybrids",
        }

        self._as_conversion = {
            "NAGGN": "other",
            "NAPAA": "other",
            "RRE-containing": "bacteriocin",
            "RiPP-like": "bacteriocin",
            "cf_fatty_acid": "fatty_acid",
            "cf_putative": "other",
            "cf_saccharide": "saccharide",
            "guanidinotides": "fused",
            "lanthipeptide-class-i": "lanthipeptide",
            "lanthipeptide-class-ii": "lanthipeptide",
            "lanthipeptide-class-iii": "lanthipeptide",
            "lanthipeptide-class-iv": "lanthipeptide",
            "lanthipeptide-class-v": "lanthipeptide",
            "lantipeptide": "lanthipeptide",
            "linaridin": "lanthipeptide",
            "lipolanthine": "lanthipeptide",
            "nrps": "NRPS",
            "otherks": "hglE-KS",
            "prodigiosin": "other",
            "pyrrolidine": "other",
            "ranthipeptide": "bacteriocin",
            "redox-cofactor": "other",
            "t1pks": "T1PKS",
            "t2pks": "T2PKS",
            "t3pks": "T3PKS",
            "thioamide-NRP": "other",
            "thioamitides": "bacteriocin",
            "transatpks": "transAT-PKS",
        }

    def get_gcf_as_classes(self, gcf, cutoff=0.5):
        """Get antismash classes for a gcf if antismash class occurs in more than <cutoff> of gcf.

        Args:
            - gcf: GCF NPLinker object
            - cutoff: float - fraction of the GCF that needs to contain the class for it to be counted
        Returns:
            List of str, the antismash classes present in the gcf
        """
        # todo: move to GCF object?
        gcf_size = len(gcf.bgcs)
        unlist_all_products = [
            product for bgc in gcf.bgcs for product in bgc.product_prediction.split(".")
        ]
        sorted_as_classes = Counter(unlist_all_products).most_common()
        # keep if in more than half of bgcs?
        cutoff = 0.5
        size_cutoff = gcf_size * cutoff
        filtered_as_classes = []
        for product in sorted_as_classes:
            if product[1] >= size_cutoff:
                filtered_as_classes.append(product[0])
        return filtered_as_classes

    def convert_as_classes(self, init_as_classes: list):
        """Convert AS classes to class names that are in scoring table.

        Args:
            - init_as_classes: list of str, the initial antismash class names
        Returns:
            List of str: converted antismash classes with _as_conversion_table
        """
        as_classes = []
        for as_class in init_as_classes:
            as_conversion = self.as_conversion.get(as_class)
            if as_conversion:
                as_classes.append(as_conversion)
            else:
                as_classes.append(as_class)
        return as_classes

    def _read_mibig_classes(self):
        """Read mibig file to dict of list {chem_id: [bgc_classes, chem_classes]}.

        Returns:
            dict(str, list(list(str))) - {chem_id: [bgc_classes, chem_classes]}
        """
        classes_dict = {}
        with open(self._mibig_classes_file) as inf:
            header = inf.readline()
            for line in inf:
                elems = line.strip().split("\t")
                chem_id = elems.pop(0)
                class_base = elems.pop(0).split(",")
                classes = [cls.partition(":")[0] for cls in class_base]
                sub_classes = [cls for cls in class_base if cls.split(":")[1]]
                as_classes = elems.pop(0).split(",")

                bgc_classes = [classes, sub_classes, as_classes]
                chem_classes = [chem_cls.split("; ") for chem_cls in elems[2:]]
                classes_dict[chem_id] = [bgc_classes, chem_classes]
        self._mibig_classes = classes_dict
        # add header info
        s_h = header.strip().split("\t")

        self._bgc_class_names = ["mibig_classes"] + s_h[1:3]
        self._chem_class_names = s_h[5:]

        return self._mibig_classes

    def _get_class_counts(self):
        """Aggregate pairwise class matrices for all compounds.

        Returns:
            Recurring defaultdict of {bgc_cat: chem_cat: bgc_c: chem_c: int}
        """

        def _rec_dd():
            """Initialises a recurring defaultdict."""
            return defaultdict(_rec_dd)

        result = _rec_dd()
        # loop through each mibig compound
        for mibig_chem_id, (bgc_classes, chem_classes) in self._mibig_classes.items():
            # get all combinations of classes for this compound
            for i, bgc_cat in enumerate(self.bgc_class_names):
                init_bgc_class = bgc_classes[i]
                if not init_bgc_class or init_bgc_class == [""]:
                    continue

                bgc_class = init_bgc_class[:]  # if no exceptions, just assign classes

                # do some cleanup for mibig classes
                if bgc_cat == "mibig_classes":
                    # group pks-nrp hybrids for MIBiG classes
                    hyb_count = len(
                        [
                            1
                            for init_bgc_c in init_bgc_class
                            if any(
                                [
                                    test in init_bgc_c.lower()
                                    for test in ["nrp", "pks", "polyketide"]
                                ]
                            )
                        ]
                    )
                    if hyb_count >= 2:
                        # if hybrid, reconstruct the bgc_class
                        bgc_class = ["PKS-NRP_Hybrids"]
                        # append other classes if there are more
                        for init_bgc_c in init_bgc_class:
                            if not any(
                                [
                                    test in init_bgc_c.lower()
                                    for test in ["nrp", "pks", "polyketide"]
                                ]
                            ):
                                bgc_class.append(init_bgc_c)

                    # replace Alkaloid with Other in bgc_class
                    bgc_class = ["Other" if bgc_c == "Alkaloid" else bgc_c for bgc_c in bgc_class]

                for j, chem_cat in enumerate(self.chem_class_names):
                    chem_class = chem_classes[j]
                    if not chem_class or chem_class == [""]:
                        continue

                    for bgc_c in bgc_class:
                        for chem_c in chem_class:
                            try:
                                result[bgc_cat][chem_cat][bgc_c][chem_c] += 1
                            except TypeError:
                                result[bgc_cat][chem_cat][bgc_c][chem_c] = 1
        self._class_count_dict = result
        return result

    def _get_scoring_tables(self):
        """Makes dicts linked to pd.DataFrame that stores counts/scores.

        The resulting dataframes (tables) are stored in _class_matches and
        _class_matches_counts.

        They should be accessed from column to row so that if you want a
        count/score from BGC class -> chem class you should query the
        _class_matches as _class_matches[bgc_class_category]\
        [chem_class_category][bgc_class_name][chem_class_name]
        """
        class_matching_tables = {}
        class_matching_counts = {}  # store the counts in df/get rid of defaultdicts
        for bgc_key, bgc_chem_counts in self._class_count_dict.items():
            for chem_key, counts in bgc_chem_counts.items():
                # init entries in dict
                if bgc_key not in class_matching_tables:
                    class_matching_tables[bgc_key] = {}
                    class_matching_counts[bgc_key] = {}
                if chem_key not in class_matching_tables:
                    class_matching_tables[chem_key] = {}
                    class_matching_counts[chem_key] = {}
                # add matching tables as DataFrames
                counts_df = pd.DataFrame.from_dict(counts)
                class_matching_tables[bgc_key][chem_key] = (
                    counts_df / counts_df.sum(axis=0)
                ).fillna(0)
                class_matching_counts[bgc_key][chem_key] = counts_df.fillna(0)
                class_matching_tables[chem_key][bgc_key] = (
                    counts_df.T / counts_df.sum(axis=1)
                ).fillna(0)
                class_matching_counts[chem_key][bgc_key] = counts_df.T.fillna(0)
        self._class_matches = class_matching_tables
        self._class_matches_counts = class_matching_counts
        return class_matching_tables

    @property
    def class_matches(self):
        return self._class_matches

    @property
    def class_matches_counts(self):
        return self._class_matches_counts

    @property
    def bgc_class_names(self):
        return self._bgc_class_names

    @property
    def chem_class_names(self):
        return self._chem_class_names

    @property
    def bigscape_mibig_conversion(self):
        return self._bigscape_mibig_conversion

    @property
    def as_conversion(self):
        return self._as_conversion



from __future__ import annotations
from collections.abc import Sequence
import pandas as pd
from nplinker.genomics import GCF
from nplinker.metabolomics import MolecularFamily
from nplinker.metabolomics import Spectrum
from nplinker.strain import StrainCollection


def get_presence_gcf_strain(gcfs: Sequence[GCF], strains: StrainCollection) -> pd.DataFrame:
    """Get the occurrence of strains in gcfs.

    The occurrence is a DataFrame with GCF objects as index and Strain objects as columns, and the
    values are 1 if the gcf occurs in the strain,  0 otherwise.
    """
    df_gcf_strain = pd.DataFrame(
        0,
        index=gcfs,
        columns=list(strains),
        dtype=int,
    )  # type: ignore
    for gcf in gcfs:
        for strain in strains:
            if gcf.has_strain(strain):
                df_gcf_strain.loc[gcf, strain] = 1
    return df_gcf_strain  # type: ignore


def get_presence_spec_strain(
    spectra: Sequence[Spectrum], strains: StrainCollection
) -> pd.DataFrame:
    """Get the occurrence of strains in spectra.

    The occurrence is a DataFrame with Spectrum objects as index and Strain objects as columns, and
    the values are 1 if the spectrum occurs in the strain, 0 otherwise.
    """
    df_spec_strain = pd.DataFrame(
        0,
        index=spectra,
        columns=list(strains),
        dtype=int,
    )  # type: ignore
    for spectrum in spectra:
        for strain in strains:
            if spectrum.has_strain(strain):
                df_spec_strain.loc[spectrum, strain] = 1
    return df_spec_strain  # type: ignore


def get_presence_mf_strain(
    mfs: Sequence[MolecularFamily], strains: StrainCollection
) -> pd.DataFrame:
    """Get the occurrence of strains in molecular families.

    The occurrence is a DataFrame with MolecularFamily objects as index and Strain objects as
    columns, and the values are 1 if the molecular family occurs in the strain, 0 otherwise.
    """
    df_mf_strain = pd.DataFrame(
        0,
        index=mfs,
        columns=list(strains),
        dtype=int,
    )  # type: ignore
    for mf in mfs:
        for strain in strains:
            if mf.has_strain(strain):
                df_mf_strain.loc[mf, strain] = 1
    return df_mf_strain  # type: ignore



from __future__ import annotations
import os
import zipfile
from os import PathLike
from pathlib import Path
from nplinker import utils
from .gnps_format import GNPSFormat
from .gnps_format import gnps_format_from_archive


class GNPSExtractor:
    """Extract files from a GNPS molecular networking archive (.zip).

    ??? info "Concept"
        [GNPS data][gnps-data]

    Four files are extracted and renamed to the following names:

    - file_mappings(.tsv/.csv)
    - spectra.mgf
    - molecular_families.tsv
    - annotations.tsv

    The files to be extracted are selected based on the GNPS workflow type,
    as described below (in the order of the files above):

    1. METABOLOMICS-SNETS
        - clusterinfosummarygroup_attributes_withIDs_withcomponentID/*.tsv
        - METABOLOMICS-SNETS*.mgf
        - networkedges_selfloop/*.pairsinfo
        - result_specnets_DB/*.tsv
    2. METABOLOMICS-SNETS-V2
        - clusterinfosummarygroup_attributes_withIDs_withcomponentID/*.clustersummary
        - METABOLOMICS-SNETS-V2*.mgf
        - networkedges_selfloop/*.selfloop
        - result_specnets_DB/.tsv
    3. FEATURE-BASED-MOLECULAR-NETWORKING
        - quantification_table*/*.csv
        - spectra/*.mgf
        - networkedges_selfloop/*.selfloop
        - DB_result/*.tsv

    Attributes:
        gnps_format: The GNPS workflow type.
        extract_dir: The path where to extract the files to.
    """

    def __init__(self, file: str | PathLike, extract_dir: str | PathLike):
        """Initialize the GNPSExtractor.

        Args:
            file: The path to the GNPS zip file.
            extract_dir: path to the directory where to extract the files to.

        Raises:
            ValueError: If the given file is an invalid GNPS archive.

        Examples:
            >>> gnps_extractor = GNPSExtractor("path/to/gnps_archive.zip", "path/to/extract_dir")
            >>> gnps_extractor.gnps_format
            <GNPSFormat.SNETS: 'METABOLOMICS-SNETS'>
            >>> gnps_extractor.extract_dir
            'path/to/extract_dir'
        """
        gnps_format = gnps_format_from_archive(file)
        if gnps_format == GNPSFormat.Unknown:
            raise ValueError(
                f"Unknown workflow type for GNPS archive '{file}'."
                f"Supported GNPS workflows are described in the GNPSFormat enum, "
                f"including such as 'METABOLOMICS-SNETS', 'METABOLOMICS-SNETS-V2' "
                f"and 'FEATURE-BASED-MOLECULAR-NETWORKING'."
            )

        self._file = Path(file)
        self._extract_path = Path(extract_dir)
        self._gnps_format = gnps_format
        # the order of filenames matters
        self._target_files = [
            "file_mappings",
            "spectra.mgf",
            "molecular_families.tsv",
            "annotations.tsv",
        ]

        self._extract()

    @property
    def gnps_format(self) -> GNPSFormat:
        """Get the GNPS workflow type.

        Returns:
            GNPS workflow type.
        """
        return self._gnps_format

    @property
    def extract_dir(self) -> str:
        """Get the path where to extract the files to.

        Returns:
            Path where to extract files as string.
        """
        return str(self._extract_path)

    def _extract(self):
        """Extract required files from archive."""
        if self._gnps_format == GNPSFormat.SNETS:
            self._extract_snets()
        elif self._gnps_format == GNPSFormat.SNETSV2:
            self._extract_snetsv2()
        elif self._gnps_format == GNPSFormat.FBMN:
            self._extract_fbmn()

    def _extract_snets(self):
        # the order of members matters
        members = [
            self._select_member(
                "clusterinfosummarygroup_attributes_withIDs_withcomponentID", ".tsv"
            ),
            self._select_member("METABOLOMICS-SNETS", ".mgf"),
            self._select_member("networkedges_selfloop", ".pairsinfo"),
            self._select_member("result_specnets_DB", ".tsv"),
        ]
        utils.extract_archive(self._file, self._extract_path, members)
        # rename the files to the expected names
        # os.renames automatically remove empty directories after renaming
        os.renames(
            self._extract_path / members[0], self._extract_path / (self._target_files[0] + ".tsv")
        )
        for member, fname in zip(members[1:], self._target_files[1:]):
            os.renames(self._extract_path / member, self._extract_path / fname)

    def _extract_snetsv2(self):
        # the order of members matters
        members = [
            self._select_member(
                "clusterinfosummarygroup_attributes_withIDs_withcomponentID", ".clustersummary"
            ),
            self._select_member("METABOLOMICS-SNETS-V2", ".mgf"),
            self._select_member("networkedges_selfloop", ".selfloop"),
            self._select_member("result_specnets_DB", ".tsv"),
        ]
        utils.extract_archive(self._file, self._extract_path, members)
        os.renames(
            self._extract_path / members[0], self._extract_path / (self._target_files[0] + ".tsv")
        )
        for member, fname in zip(members[1:], self._target_files[1:]):
            os.renames(self._extract_path / member, self._extract_path / fname)

    def _extract_fbmn(self):
        # there might be two folders for quantification table
        # "quantification_table_reformatted" and "quantification_table"
        try:
            quantification_table_member = self._select_member(
                "quantification_table_reformatted", ".csv"
            )
        except ValueError:
            quantification_table_member = self._select_member("quantification_table", ".csv")

        # the order of members matters
        members = [
            quantification_table_member,
            self._select_member("spectra", ".mgf"),
            self._select_member("networkedges_selfloop", ".selfloop"),
            self._select_member("DB_result", ".tsv"),
        ]
        utils.extract_archive(self._file, self._extract_path, members)
        os.renames(
            self._extract_path / members[0], self._extract_path / (self._target_files[0] + ".csv")
        )
        for member, fname in zip(members[1:], self._target_files[1:]):
            os.renames(self._extract_path / member, self._extract_path / fname)

    def _select_member(self, prefix: str, suffix: str) -> str:
        """Helper function to extract files matching a prefix and suffix from the archive."""
        with zipfile.ZipFile(self._file) as zf:
            member_list = [
                member
                for member in zf.namelist()
                if member.startswith(prefix) and member.endswith(suffix)
            ]
            if len(member_list) != 1:
                raise ValueError(
                    f"Expected exactly one file matching pattern '{prefix}*{suffix}'"
                    f"in archive '{self._file}', but found {len(member_list)}."
                )
        return member_list[0]



from __future__ import annotations
from functools import cached_property
from typing import TYPE_CHECKING
import numpy as np
from nplinker.strain import Strain
from nplinker.strain import StrainCollection


if TYPE_CHECKING:
    from .molecular_family import MolecularFamily


class Spectrum:
    """Class to model MS/MS Spectrum.

    Attributes:
        id: the spectrum ID.
        mz: the list of m/z values.
        intensity: the list of intensity values.
        precursor_mz: the m/z value of the precursor ion.
        precursor_charge: the charge of the precursor ion.
        rt: the retention time in seconds.
        metadata: the metadata of the spectrum, i.e. the header information in the MGF file.
        gnps_annotations: the GNPS annotations of the spectrum.
        gnps_id: the GNPS ID of the spectrum.
        strains: the strains that this spectrum belongs to.
        family: the molecular family that this spectrum belongs to.
        peaks: 2D array of peaks, each row is a peak of (m/z, intensity) values.
    """

    def __init__(
        self,
        id: str,
        mz: list[float],
        intensity: list[float],
        precursor_mz: float,
        precursor_charge: int,
        rt: float = 0,
        metadata: dict | None = None,
    ) -> None:
        """Initialize the Spectrum.

        Args:
            id: the spectrum ID.
            mz: the list of m/z values.
            intensity: the list of intensity values.
            precursor_mz: the precursor m/z.
            precursor_charge: the charge of the precursor ion.
            rt: the retention time in seconds. Defaults to 0.
            metadata: the metadata of the spectrum, i.e. the header information
                in the MGF file.
        """
        self.id = id
        self.mz = mz
        self.intensity = intensity
        self.precursor_mz = precursor_mz
        self.precursor_charge = precursor_charge
        self.rt = rt
        self.metadata = metadata or {}

        self.gnps_annotations: dict = {}
        self.gnps_id: str | None = None
        self.strains: StrainCollection = StrainCollection()
        self.family: MolecularFamily | None = None

    def __str__(self) -> str:
        return f"Spectrum(id={self.id}, #strains={len(self.strains)})"

    def __repr__(self) -> str:
        return str(self)

    def __eq__(self, other) -> bool:
        if isinstance(other, Spectrum):
            return self.id == other.id and self.precursor_mz == other.precursor_mz
        return NotImplemented

    def __hash__(self) -> int:
        return hash((self.id, self.precursor_mz))

    def __reduce__(self) -> tuple:
        """Reduce function for pickling."""
        return (
            self.__class__,
            (
                self.id,
                self.mz,
                self.intensity,
                self.precursor_mz,
                self.precursor_charge,
                self.rt,
                self.metadata,
            ),
            self.__dict__,
        )

    @cached_property
    def peaks(self) -> np.ndarray:
        """Get the peaks, a 2D array with each row containing the values of (m/z, intensity)."""
        return np.array(list(zip(self.mz, self.intensity)))

    def has_strain(self, strain: Strain) -> bool:
        """Check if the given strain exists in the spectrum.

        Args:
            strain: `Strain` object.

        Returns:
            True when the given strain exist in the spectrum.
        """
        return strain in self.strains



import itertools
import logging
from nplinker.genomics.bgc import BGC
from nplinker.genomics.gcf import GCF
from nplinker.metabolomics import MolecularFamily
from nplinker.scoring.abc import ScoringBase
from nplinker.scoring.rosetta.rosetta import Rosetta
from .link_graph import LinkGraph
from .score import Score
from .scoring_method import ScoringMethod


logger = logging.getLogger(__name__)


class RosettaScoring(ScoringBase):
    name = ScoringMethod.ROSETTA.value
    ROSETTA_OBJ = None

    def __init__(self, npl):
        super().__init__(npl)
        self.bgc_to_gcf = True

        self.spec_score_cutoff = 0.0
        self.bgc_score_cutoff = 0.0

    @classmethod
    def setup(cls, npl):
        """Setup the Rosetta object and run the scoring algorithm.

        This method is only called once to setup the Rosetta object.
        """
        logger.info("RosettaScoring setup")
        cls.ROSETTA_OBJ = Rosetta(npl, ignore_genomic_cache=False)
        ms1_tol = Rosetta.DEF_MS1_TOL
        ms2_tol = Rosetta.DEF_MS2_TOL
        score_thresh = Rosetta.DEF_SCORE_THRESH
        min_match_peaks = Rosetta.DEF_MIN_MATCH_PEAKS

        ms1_tol, ms2_tol, score_thresh, min_match_peaks = RosettaScoring._init_from_config(
            npl.config
        )

        cls.ROSETTA_OBJ.run(npl.spectra, npl.bgcs, ms1_tol, ms2_tol, score_thresh, min_match_peaks)
        logger.info("RosettaScoring setup completed")

    @staticmethod
    def _init_from_config(config):
        """Allow overridding params via config file."""
        if "scoring" in config and "rosetta" in config["scoring"]:
            rc = config["scoring"]["rosetta"]
            ms1_tol = rc.get("ms1_tol", Rosetta.DEF_MS1_TOL)
            ms2_tol = rc.get("ms2_tol", Rosetta.DEF_MS2_TOL)
            score_thresh = rc.get("score_thresh", Rosetta.DEF_SCORE_THRESH)
            min_match_peaks = rc.get("min_match_peaks", Rosetta.DEF_MIN_MATCH_PEAKS)

        return ms1_tol, ms2_tol, score_thresh, min_match_peaks

    def _include_hit(self, hit):
        if (
            hit.spec_match_score < self.spec_score_cutoff
            or hit.bgc_match_score < self.bgc_score_cutoff
        ):
            return False

        return True

    def _insert_result_gen(self, results, src, hit):
        if src not in results:
            results[src] = {}
        # Rosetta can produce multiple "hits" per link
        if hit.spec in results[src]:
            original_data = results[src][hit.spec].data(self)
            results[src][hit.spec].set_data(self, original_data + [hit])
        else:
            results[src][hit.spec] = Score(name=self.name, value=[hit], parameter=self._params)

        return results

    def _insert_result_met(self, results, spec, target, hit):
        if spec not in results:
            results[spec] = {}
        # Rosetta can produce multiple "hits" per link
        if target in results[spec]:
            original_data = results[spec][target].data(self)
            results[spec][target].set_data(self, original_data + [hit])
        else:
            results[spec][target] = Score(name=self.name, value=[hit], parameter=self._params)

        return results

    def get_links(self, *objects, **parameters):
        # TODO: replace some attributes with parameters
        self._params = parameters

        self._validate_inputs(objects)

        if isinstance(objects[0], GCF):
            # assume user wants to use all BGCs from these GCFs
            bgcs = list(set(itertools.chain.from_iterable(x.bgcs for x in objects)))
            logger.info(
                "RosettaScoring got {} GCFs input, converted to {} BGCs".format(
                    len(objects), len(bgcs)
                )
            )
            objects = bgcs

        # list of RosettaHit objects which satisfy the current cutoffs
        ro_hits = list(
            filter(lambda hit: self._include_hit(hit), RosettaScoring.ROSETTA_OBJ._rosetta_hits)
        )

        # TODO this might need to be faster
        results = {}
        if isinstance(objects[0], BGC):
            results = self._collect_results_bgc(objects, ro_hits, results)
        else:  # Spectrum
            results = self._collect_results_spectra(objects, ro_hits, results)
        logger.debug(f"RosettaScoring found {len(results)} results")

        lg = LinkGraph()
        for src, links in results.items():
            for target, score in links.items():
                lg.add_link(src, target, score)
        return lg

    def _collect_results_spectra(self, objects, ro_hits, results):
        for spec in objects:
            for hit in ro_hits:
                if spec.id == hit.spec.id:
                    if not self.bgc_to_gcf:
                        # can use the BGC directly
                        results = self._insert_result_met(results, spec, hit.bgc, hit)
                    else:
                        # if we want the results to contain GCFs instead, need
                        # to iterate over the set of bgc.parents (most will only
                        # have one, hybrid BGCs will have more)
                        for gcf in hit.bgc.parents:
                            results = self._insert_result_met(results, spec, gcf, hit)
        return results

    def _collect_results_bgc(self, objects, ro_hits, results):
        for bgc in objects:
            for hit in ro_hits:
                if bgc.id == hit.bgc.id:
                    if not self.bgc_to_gcf:
                        # can use the BGC directly
                        results = self._insert_result_gen(results, bgc, hit)
                    else:
                        # if we want the results to contain GCFs instead, need
                        # to iterate over the set of bgc.parents (most will only
                        # have one, hybrid BGCs will have more)
                        for gcf in bgc.parents:
                            results = self._insert_result_gen(results, gcf, hit)
        return results

    def _validate_inputs(self, objects):
        """Enforce constraint that the list must contain a set of identically typed objects."""
        if not all(isinstance(x, type(objects[0])) for x in objects):
            raise Exception("RosettaScoring: uniformly-typed list of objects is required")

        if isinstance(objects[0], MolecularFamily):
            raise Exception("RosettaScoring requires input type Spectrum (found MolecularFamily)")

    def format_data(self, data):
        # TODO
        return f"{len(data)} hits"

    def sort(self, objects, reverse=True):
        # TODO
        return objects
